"""
Synonyms and flowables for the WorldSteel XLSX file
"""
from synonym_dict import MergeError
from synonym_dict.synonym_dict import ParentNotFound

import re


discovered_synonyms = [
    ('Zinc(ii)', 'Zinc, ion'),
    ('Copper(ii)', 'Copper (II)'),
    ('Vanadium(v)', 'Vanadium, ion'),
    ('Nickel(ii)', 'Nickel, ion'),
    ('Silver(i)', 'Silver, ion'),
    ('Tin(ii)', 'Tin, ion'),
    ('sulfur oxides', 'Sulphur oxides'),
    ('sulfur dioxide', 'sulphur dioxide'),
    ('sulfur trioxide', 'sulphur trioxide'),
    ('sulfuric acid', 'sulphuric acid'),
    ('hydrogen sulfide', 'hydrogen sulphide'),
    ('pm2.5', 'Particulate matter, ≤ 2.5μm', 'Dust (PM2.5)'),
    ('pm10', 'Particulate matter, ≤ 10μm'),
    ('pm10', 'Particulate matter', 'Dust (> PM10)'),
    ('pm10', 'Particulate matter, > 2.5μm and ≤ 10μm', 'Dust (PM2.5 - PM10)'),
    ('Ethane, 1,1,2-trichloro-1,2,2-trifluoro-, CFC-113', 'cfc-113'),
    ('Ethane, 1,2-dichloro-1,1,2,2-tetrafluoro-, CFC-114', 'cfc-114'),
    ('Methane, bromodifluoro-, Halon 1201', 'halon 1201'),
    ('Methane, bromochlorodifluoro-, Halon 1211', 'halon 1211'),
    ('Ethane, 1-chloro-1,1-difluoro-, HCFC-142b', 'hcfc-142b'),
    ('Ethane, chloropentafluoro-, CFC-115', 'cfc-115'),
    ('Ethane, 1,2-dibromotetrafluoro-, Halon 2402', 'halon 2402'),
    ('Methane, bromotrifluoro-, Halon 1301', 'halon 1301'),
    ('Methane, dibromodifluoro-, Halon 1202', 'halon 1202'),
    ('Ethane, 1,1-dichloro-1-fluoro-, HCFC-141b', 'hcfc-141b'),
    ('Ethane, 1-chloro-1,1-difluoro-, HCFC-142b', 'hcfc-142b'),
    ('Ethane, 2,2-dichloro-1,1,1-trifluoro-, HCFC-123', 'hcfc-123'),
    ('chlorotetrafluoroethane', 'Ethane, 2-chloro-1,1,1,2-tetrafluoro-, HCFC-124', 'hcfc-124', 'R 124 (chlorotetrafluoroethane)'),
    ('Methane, chlorodifluoro-, HCFC-22', 'hcfc-22'),
    ('460-73-1', 'R 245fa (1,1,1,3,3-Pentafluoropropane)'),
    ('354-33-6', 'R 125 (pentafluoroethane)'),
    ('430-66-0', 'R 143 (trifluoroethane)'),
    ('76-16-4', 'R 116 (hexafluoroethane)'),
    ('Methane, tetrachloro-, R-10', 'tetrachloromethane', 'Carbon tetrachloride (tetrachloromethane)'),
    ('Methane, tetrafluoro-, R-14', 'tetrafluoromethane'),
    ('Methane, trichlorofluoro-, CFC-11', 'CFC-11'),
    ('75-10-5', 'R 32 (difluoromethane)', 'difluoromethane'),
    ('75-46-7', 'trifluoromethane', 'R 23 (trifluoromethane)'),
    ('75-09-2', 'methylene chloride', 'Dichloromethane (methylene chloride)'),
    ('75-43-4', 'R 21 (Dichlorofluoromethane)', 'Dichlorofluoromethane'),
    ('74-87-3', 'Chloromethane (methyl chloride)'),
    ('811-97-2', 'tetrafluoroethane'),
    ('75-37-6', 'R 152a (difluoroethane)'),
    ('Propane, 1,3-dichloro-1,1,2,2,3-pentafluoro-, HCFC-225cb', 'hcfc-225cb', 'hcfc225cb', 'hcfc 225cb'),
    ('Propane, 3,3-dichloro-1,1,1,2,2-pentafluoro-, HCFC-225ca', 'hcfc-225ca', 'hcfc225ca', 'hcfc 225ca'),
    ('Cyclohexane, pentyl-', 'pentyl cyclohexane'),
    ('75-65-0', 't-Butyl alcohol'),
    ('622-96-8', 'para-Ethyltoluene'),
    ('106-42-3', 'para-xylene'),
    ('75-00-3', 'monochloroethane'),
    ('611-14-3', 'ortho-ethyltoluene'),
    ('Nitric oxide', 'nitrogen monoxide'),
    ('Biological oxygen demand', 'biological oxygen demand (BOD)'),
    ('Chemical oxygen demand', 'chemical oxygen demand (COD)'),
    ('Nitrous oxide', 'Nitrous oxide (laughing gas)'),
    ('chloroform', 'Trichloromethane (chloroform)'),
    ('Water, river', 'River water', 'River water, regionalized'),
    ('Cooling water to river', 'Cooling water to river, regionalized'),
    ('Water, fresh', 'Fresh water'),
    ('Water, ground', 'Ground water', 'Ground water, regionalized'),
    ('Water, lake', 'Lake water', 'Lake water, regionalized'),
    ('Lake water to turbine', 'Lake water to turbine, regionalized'),
    ('River water to turbine', 'River water to turbine, regionalized'),
    ('Turbined water to river', 'Turbined water to river, regionalized'),
    ('Processed water to river', 'Processed water to river, regionalized')
]


WS_skips = {'unspecified', 'unspec.', 'biotic', 'deposited', 'total N', 'N-compounds', 'aviation', 'dissolved',
            'suspended', 'process water', 'waste water, treated', 'unspecified, as N', 'land use change',
            'atmospheric nitrogen', 'underground deposit', 'PAH, unspec.', 'unspecific', 'isomers',
            'evapotranspiration', 'general', 'as total N', 'carcinogen', 'peat oxidation'}


'''
The following regexp matches names like "some name (some synonym)" where groups(1) = ('some name', 'some synonym')
'''
exp = re.compile(r'^([^(]+)\s\(([^)]+)\)$')


def ws_syn_finder(_cat, iters):
    s = set()
    for ws_flow in iters:
        if _cat.lcia_engine[ws_flow.context].sense != 'Sink':
            continue
        m = exp.match(ws_flow.name)
        if bool(m):
            n, p = m.groups(1)
            if n == 'Dust':
                continue
            if p not in WS_skips and len(p) > 6:
                ps = tuple(filter(lambda x: x not in WS_skips, p.split('; ')))
                s.add((n, ws_flow.name, *ps))
    return s


def _known_unknown(_cat, *terms):
    uk = set(terms)
    for t in terms:
        try:
            k = _cat.lcia_engine.get_flowable(t)
            uk.remove(t)
            return k.name, uk
        except KeyError:
            pass


def _worldsteel_synonyms(_cat, syns):
    try:
        _cat.lcia_engine.add_terms('flow', *syns)
    except MergeError:
        k, uk = _known_unknown(_cat, *syns)
        if k:
            try:
                _cat.lcia_engine.merge_flowables(k, *uk)
                print('merged %s' % (syns,))
            except ParentNotFound:
                print('wtf %s' % (syns,))
        else:
            print('could not merge %s' % (syns,))


def worldsteel_flowables(cat, ws):
    """
    Attempts to add flowables with the format "One name, comma-separated (alternate name)",
    with a large set of exclusions.  Adds 'one name, comma-separated' and 'alternative name' as synonyms.
    :param cat: a catalog
    :param ws: a query whose flows() have names matching the above format
    :return:
    """
    # worldsteel flowables
    ws_syns = ws_syn_finder(cat, ws.flows())

    for syns in ws_syns:
        _worldsteel_synonyms(cat, syns)



import os

from antelope_reports.model_runner.lca_model_runner import tabularx_ify
from antelope_reports.charts.pos_neg import PosNegCompareError


class ResultsWriter(object):

    pdf = True

    @property
    def unit_output(self):
        """
        This is the unit-scale case-by-case output as a long table
        :return:
        """
        return os.path.join(self.lca_path, '%s-unit-models%s.csv' % (self.scope, self.suffix))

    @property
    def unit_csv(self):
        """
        Summary table (stage x lcia) in csv format
        :return:
        """
        return os.path.join(self.lca_path, '%s-unit-tabular%s.csv' % (self.scope, self.suffix))

    @property
    def unit_table(self):
        """
        Summary table (stage x lcia) in tex format
        :return:
        """
        return os.path.join(self.figs_path, '%s-unit-bar%s.tex' % (self.scope, self.suffix))

    @property
    def full_output(self):
        """
        Provincial-scale output as a long table for all scenarios
        :return:
        """
        return os.path.join(self.lca_path, '%s-lca%s.csv' % (self.scope, self.suffix))

    def year_csv(self, year):
        """
        Provincial-scale summary table (stage x lcia) for a single year, csv format
        :param year:
        :return:
        """
        return os.path.join(self.lca_path, '%s-lca-%s-tabular%s.csv' % (self.scope, year, self.suffix))

    def year_table(self, year):
        """
        Provincial-scale summary table (stage x lcia) for a single year, tex format
        :param year:
        :return:
        """
        return os.path.join(self.figs_path, '%s-lca-%s-bar%s.tex' % (self.scope, year, self.suffix))

    def pos_neg_eps(self, scenario):
        return os.path.join(self.figs_path, '%s-pos-neg-%s%s.%s' % (self.scope, scenario, self.suffix, self._or_eps))

    def pos_neg_tex(self, scenario):
        return os.path.join(self.figs_path, '%s-pos-neg-%s%s.tex' % (self.scope, scenario, self.suffix))

    def system_diagram(self, *scenarios):
        scen = '-'.join(scenarios)
        return os.path.join(self.figs_path, '%s-system-diagram_%s.fig' % (self.scope, scen))

    def _check_output_dir(self):
        if not os.path.exists(self.lca_path):
            os.makedirs(self.lca_path, exist_ok=True)
        if not os.path.exists(self.figs_path):
            os.makedirs(self.figs_path, exist_ok=True)

    @property
    def lca_path(self):
        return os.path.join(self.lca_root, self.scope)

    @property
    def figs_path(self):
        return os.path.join(self.figs_root, self.scope)

    @property
    def suffix(self):
        if self.case is None:
            return ''
        return '-%s' % self.case

    def __init__(self, path, scope, lca_path='lca_output', figs_path='lca_figs', case=None):
        """

        :param path:
        :param scope:
        :param lca_path:
        :param figs_path:
        :param case:
        """

        self.scope = scope
        self.case = case

        self.lca_root = os.path.abspath(os.path.join(path, lca_path))
        self.figs_root = os.path.abspath(os.path.join(path, figs_path))

        self._check_output_dir()

    def generate_csv(self, catra, **kwargs):
        catra.results_to_csv(self.full_output, **kwargs)

    def generate_year_output(self, study_year, year, stage_order):
        # generate_pos_neg_compare(study_year)
        study_year.scenario_detail_tbl(year, filename=self.year_csv(year), column_order=stage_order)

        study_year.results_to_tex(self.year_table(year), scenario=year, column_order=stage_order, format='%.2e',
                                  sort_column=0)

    def generate_unit_output(self, study, scenario_order=None):
        """
        set display quantities via runner
        :param study: LcaModelRunner instance
        :param scenario_order:
        :return:
        """
        study.results_to_csv(self.unit_output, scenarios=scenario_order)

        study.results_to_tex(self.unit_table, column_order=scenario_order, format='%.2e', sort_column=0)

    def pos_neg_chart(self, runner, scenario, qs=None, table=True):
        if qs is None:
            qs = list(runner.quantities)
        pn = PosNegCompareError(*(runner.sens_result(scenario, q) for q in qs), filename=self.pos_neg_eps(scenario))
        if table:
            # multicolumn=False is no longer supported
            tabularx_ify(pn.dataframe, column_format='ll *{3}X',
                         filename=self.pos_neg_tex(scenario))  # multicolumn=False

    @property
    def _or_png(self):
        if self.pdf:
            return 'pdf'
        return 'png'

    @property
    def _or_eps(self):
        if self.pdf:
            return 'pdf'
        return 'png'

    def __enter__(self):
        return self

    def __exit__(self, *args):
        print('Exiting %s context' % self.__class__.__name__)
        print(args)
        pass



from .conventions import activity_model_ref, transport_model, logistics_summary_ref
from .lca_model import LcaModel
from .lca_study import build_study_foreground, scenario_knobs
from .markets import make_routes, build_market_mix, market_mixers






class FlowComparator(object):
    """
    A tool for comparing the overlap and difference between different families of flows.
    Use cases: LCIA indicators; archives

    usage:
    > fc = FlowComparator(cat)
    > fc.sort_flowables(q1, q2)

    where q1, q2 are queries: could be quantities, archives, or lists of fragments or exchanges
    maps each item's flows to recognized flowables
    returns 3 sets: flowables unique to q1, common to both, unique to q2
    records: terms encountered in queries that are not recognized flowables

    """

    def __init__(self, tm):
        self._tm = tm
        self._unknown_flows = dict()

    def _lcia_get_flowable(self, _flowable, _origin):
        if _origin not in self._unknown_flows:
            self._unknown_flows[_origin] = set()
        try:
            return self._tm.get_flowable(_flowable)
        except KeyError:
            self._unknown_flows[_origin].add(_flowable)
            return _flowable

    def _map_item(self, item):
        """

        :param item:
        :return:
        """
        if hasattr(item, 'entity_type'):
            if item.entity_type == 'characterization':
                fb = item.flowable
                og = item.origin
            elif item.entity_type == 'exchange':
                fb = item.flow.name
                og = item.flow.origin
            elif item.entity_type == 'flow':
                fb = item.name
                og = item.origin
            elif item.entity_type == 'fragment':
                fb = item.flow.name
                og = item.origin
            else:
                raise TypeError(item)
        else:
            fb = str(item)
            og = None
        return fb, self._lcia_get_flowable(fb, og)

    def distinct_flowables(self, _q):
        if hasattr(_q, 'factors'):
            fs = (self._map_item(k) for k in _q.factors())
            _n = _q.name
        elif hasattr(_q, 'flows'):
            fs = (self._map_item(k) for k in _q.flows())
            if hasattr(_q, 'origin'):
                _n = _q.origin
            elif hasattr(_q, 'ref'):
                _n = _q.ref
            else:
                _n = _q.__class__.__name__
        else:
            fs = (self._map_item(k) for k in _q)
            _n = _q.__class__.__name__

        ff, ss = zip(*fs)
        bb = set(ss)

        print('%s: %d factors; %d names; %d flowables' % (_n, len(ff), len(set(ff)), len(bb)))

        return bb

    def sort_flowables(self, _q1, _q2):
        """
        returns lcia-engine flowables in 3 sets: distinct to _q1, common to both, distinct to _q2

        Three steps:
         - secure list of items
         - map items to strings (flagging unrecognized strings
         - map strings to flowables
        """

        _s1 = self.distinct_flowables(_q1)
        _s2 = self.distinct_flowables(_q2)
        _u = _s1.intersection(_s2)
        _d1 = _s1.difference(_u)
        _d2 = _s2.difference(_u)
        print('only a: %d ,common: %d , only b: %d' % (len(_d1), len(_u), len(_d2)))
        return _d1, _u, _d2

    @property
    def seen_origins(self):
        return list(self._unknown_flows.keys())

    def unknown_flowables(self, origin):
        return sorted(self._unknown_flows.get(origin, []))



from .lca_model_runner import LcaModelRunner
from .scenario_runner import ScenarioRunner
from .sens_runner import SensitivityRunner
from .results_writer import ResultsWriter

import pandas as pd


def get_stage_name(_ff):
    try:
        _entity = _ff.fragment
    except AttributeError:
        _entity = _ff
    tries = ('stage_name', 'stagename', 'stage', 'group')
    for i in tries:
        sn = _entity.get(i, None)
        if sn:
            return sn
    return 'undefined'


def get_child_property(_ff, property_name, default=None, concat=None):
    """
    Utility function to retrieve a given property name from inside a fragment flow's anchor
    :param _ff:
    :param property_name:
    :param default:
    :param concat:
    :return:
    """
    if _ff.term.is_context:
        return _ff.term.term_node.name
    g = _ff.term.term_node.get(property_name)
    if g is not None and g != '':
        return g
    if hasattr(_ff.term.term_node, 'child_flows'):
        gs = list(filter(None, (f.get(property_name) for f in _ff.term.term_node.child_flows)))
        if len(gs) > 0:
            if concat:
                return concat.join(gs)
            else:
                return gs[0]
        else:
            return default


def get_top_level_flow(_ff, arg='Name', default='', tops=()):
    """
    Returns the name of the topmost child flow whose parent is either the study model or in the list of tops
    :param _ff:
    :param arg:
    :param default:
    :param tops: set/tuple/etc of fragments whose child flows should be named
    :return:
    """
    if _ff.superfragment and _ff.superfragment.fragment.top() not in tops:
        return get_top_level_flow(_ff.superfragment, arg=arg, default=default)
    elif _ff.superfragment is None:
        return _ff.fragment.get(arg, default)
    return _ff.superfragment.fragment.get(arg, default)


def lcias_to_dataframe(*ress, key=None):
    """
    Concatenate a set of LCIA results into a dataframe, quantities in column x components in row,
    using c.name
    :param ress:
    :param key: executable name extractor for *component* (not entity) (default getattr(c, 'name'))
    :return:
    """
    if key is None:
        key = lambda x: getattr(x, 'name')

    def _single_qty(r):
        return pd.DataFrame(((key(c), c.cumulative_result * 100 / r.total()) for c in r.components()),
                            columns=('Stage', r.quantity['indicator'])).set_index('Stage')

    return pd.concat((_single_qty(k) for k in ress), axis=1)



from scipy import stats
from numpy.random import default_rng
from math import exp, sqrt

import logging


class Randomizer:
    """
    An abstract class, meant to generate unit-scaled random numbers to be multiplied by static means to generate
    distributions.
    """
    _engine = None

    @property
    def engine(self):
        return self._engine

    @property
    def rv(self):
        return self.engine.rvs()

    def value(self, scale):
        return scale * self.rv


class UniformRandomizer(Randomizer):

    def __init__(self, low, high):
        """
        The stats.uniform distribution is specified as a "low" and a "width"

        :param low:
        :param high:
        """
        if low >= high:
            raise ValueError('Invalid bounds %g -> %g' % (low, high))
        if low > 1.0 or high < 1.0:
            logging.warning('1.0 is not in bounds for this uniform randomizer')

        self.params = (low, high)
        self._engine = stats.uniform(low, high - low)


class LognormalRandomizer(Randomizer):

    def __init__(self, shape):
        """
        :param shape: the shape parameter for the lognormal distribution. This is dimensionless and should not depend
        on the mean.  The coefficient of variation is derived rom the shape parameter.

        """
        self.params = (shape,)
        self._engine = stats.lognorm(shape)

    @property
    def cv(self):
        """
        The coefficient of variation for the underlying uniform distribution
        """
        shape = self.params[0]
        return sqrt(exp(shape ** 2) - 1)


class DirichletRandomizer(Randomizer):
    """
    A Dirichlet distribution is a way of picking points across a simplex (that sum to a constant value).
    It returns a vector of random values equal to the number of input parameters.  Those parameters should
    be given weights relative to each other to bias the result, and relative to 1 to increase the concentration
    around the mean.  e.g. in the default parameterization of [1, 1, 1], all points in the simplex will be equally
    likely and they will all be centered around 1/3.  In a parameterization of [8, 1, 1], the first value will
    be in the vicinity of 0.8 the size of the others, but they will still be somewhat uniformly concentrated.
    Whereas in a parameterization of [0.8, 0.1, 0.1] the first value will be in the vicinity of 0.8, but the values
    will be biased toward the edges.
    """
    def __init__(self, *params):
        self.params = list(params)
        if len(self.params) < 2:
            logging.warning("Trivial Dirichlet distribution (N=%d(" % len(self.params))
        self._engine = default_rng()

    @property
    def rv(self):
        return self.engine.dirichlet(self.params)

    def value(self, scale):
        return [scale * k for k in self.rv]



from .lca_model_runner import LcaModelRunner
from pandas import ExcelWriter


class MultiModelRunner(LcaModelRunner):
    def __init__(self, *frags, **kwargs):
        super(MultiModelRunner, self).__init__(**kwargs)
        self._frags = dict()

        for frag in frags:
            self.add_fragment(frag)

    def add_fragment(self, frag, name=None):
        """

        :param frag:
        :param name:
        :return:
        """
        if name is None:
            name = frag.name
        if name not in self._scenarios:
            self.add_scenario(name)
        self._frags[name] = frag

    def _run_scenario_lcia(self, scenario, lcia, **kwargs):
        frag = self._frags[scenario]
        return frag.fragment_lcia(lcia, **kwargs)

    def write_to_xlsx(self, xlsx_name, details=False):
        """
        This is presently a big fat ugly mess and does not satisfy basic requirements because it does not specify
        the functional unit of each model.  We need to add metadata to this...
        :param xlsx_name:
        :param details:
        :return:
        """
        self._fmt = None
        xlw = ExcelWriter(xlsx_name)
        pd = self.scenario_summary_tbl()
        pd.to_excel(xlw, sheet_name='Summary')
        if details:
            for s in self.scenarios:
                self.scenario_detail_tbl(s).to_excel(xlw, sheet_name=s)
        xlw.save()



import matplotlib.pyplot as plt
import matplotlib.patches as mpatches

from .base import standard_labels, label_vbar, prefab_colors, net_color, wrap


def spread_bars(ax, data, color_gen, hi=None, lo=None, y_lim=None, barwidth=0.65, x_offset=0, labels=True, **kwargs):
    """

    :param ax:
    :param data:
    :param color_gen: iterator that yields color for data points
    :param hi: data vector for high errorbars (passed on to pyplot)
    :param lo: data vector for low errorbars (passed on to pyplot)
    :param y_lim: y axis limits, if to be specified
    :param barwidth: [0.65]
    :param x_offset: [0]
    :param labels: [True] whether to label the top of each bar
    :return:
    """
    # x = [i - barwidth/2 + x_offset for i in range(len(data))]
    x = [i + x_offset for i in range(len(data))]

    kwargs['ecolor'] = (0, 0, 0)
    kwargs['width'] = barwidth
    kwargs['color'] = [k for k in color_gen]

    if hi is None:
        hi_d = [0] * len(data)
    else:
        hi_d = [hi[i] - data[i] for i, _ in enumerate(data)]
    if lo is None:
        lo_d = [0] * len(data)
    else:
        lo_d = [data[i] - lo[i] for i, _ in enumerate(data)]

    kwargs['yerr'] = [lo_d, hi_d]

    patches = ax.bar(x, data, **kwargs)
    if labels:
        for i, patch in enumerate(patches):
            if data[i] != 0:
                label_vbar(patch, data[i], sep=hi_d[i])

    ax.set_xlim(-0.5, len(data)-0.5)
    if y_lim is not None:
        ax.set_ylim(y_lim)

    ax.plot((-0.5, len(data)), (0, 0), 'k')
    # _open_ylims(ax)


def spread_contrib_figure(results, stages, colors=None, scenarios=None, results_hi=None, results_lo=None,
                          match_y=False):
    """

    :param results: m-array of queryable results
    :param stages: k-array of stages
    :param colors: should match k
    :param scenarios: should match m (used for subtitles)
    :param results_hi: either None or an m-array of high values
    :param results_lo: either None or an m-array of low values
    :param match_y: [False] if True, set all axes to have the same [maximal] y limits
    :return:
    """
    if not isinstance(results, list):
        # assume that all the m-arrays are non-arrays
        results = [results]
        results_hi = [results_hi]
        results_lo = [results_lo]
        scenarios = [scenarios]

    if len(results) == 1:
        f = plt.figure(figsize=(6, 5))
        wid = 1

    else:
        height = 5 * round(len(results)/2)
        f = plt.figure(figsize=(12, height))
        wid = 2

    if colors is None:
        colors = (0.1, 0.6, 0.7)

    if results_hi is None:
        results_hi = [None] * len(results)

    if results_lo is None:
        results_lo = [None] * len(results)

    ax = []

    for i, result in enumerate(results):
        ax.append(f.add_subplot(len(results), wid, i + 1))
        data = result.contrib_query(stages)
        if results_hi[i] is None:
            hi = data
        else:
            hi = results_hi[i].contrib_query(stages)

        if results_lo[i] is None:
            lo = data
        else:
            lo = results_lo[i].contrib_query(stages)

        spread_bars(ax[-1], data, colors, lo=lo, hi=hi)
        standard_labels(ax[-1], stages)

        t = result.quantity['Indicator']

        if scenarios is None:
            ax[-1].set_title(t, fontsize=14)
        else:
            ax[-1].set_title('%s\n%s' % (t, scenarios[i]), fontsize=14)

    if match_y:
        y_lo = 0
        y_hi = 0
        for x in ax:
            (y0, y1) = x.get_ylim()
            if y0 < y_lo:
                y_lo = y0
            if y1 > y_hi:
                y_hi = y1
        for x in ax:
            x.set_ylim(y_lo, y_hi)

    return f


def spread_scenario_compare(ax, results, stages, colors=None, scenarios=None, results_hi=None, results_lo=None,
                            net=False, labels=False, legend=True):
    """

    :param ax: axes already created
    :param results:
    :param stages:
    :param colors:
    :param scenarios:
    :param results_hi:
    :param results_lo:
    :param net:
    :return:
    """
    barwidth = 0.8  # use most of the space for multiple scenarios

    if not isinstance(results, list):
        # assume that all the m-arrays are non-arrays
        results = [results]
        results_hi = [results_hi]
        results_lo = [results_lo]
        scenarios = [scenarios]
        barwidth = 0.65

    m = len(results)

    barwidth /= m

    x_0 = -(m + 1) * barwidth / 2

    if colors is None:
        colors = prefab_colors[:len(results)]

    if results_hi is None:
        results_hi = [None] * len(results)

    if results_lo is None:
        results_lo = [None] * len(results)

    if scenarios is None:
        scenarios = ['S%d' % (i+1) for i, _ in enumerate(results)]

    unit = list(set([(r.quantity['Indicator'], r.quantity.unit) for r in results]))
    if len(unit) > 1:
        print('Warning: multiple units found in result sets!!')

    for i, r in enumerate(results):
        x_0 += barwidth

        data = r.contrib_query(stages)
        if net:
            data.append(r.total())

        if results_hi[i] is None:
            hi = data
        else:
            hi = results_hi[i].contrib_query(stages)
            if net:
                hi.append(results_hi[i].total())

        if results_lo[i] is None:
            lo = data
        else:
            lo = results_lo[i].contrib_query(stages)
            if net:
                lo.append(results_lo[i].total())

        mycolors = (colors[i],) * len(stages)
        if net:
            mycolors += (net_color,)

        spread_bars(ax, data, mycolors, lo=lo, hi=hi, barwidth=barwidth, x_offset=x_0, labels=labels, align='edge')

    # graph general parts
    ax.plot(ax.get_xlim(), (0, 0), 'k', linewidth=0.8)

    if max([len(l) for l in stages]) > 12:
        lbls = ['\n'.join(wrap(l, 25)) for l in stages]
        rotation = 90
    else:
        lbls = [l for l in stages]
        rotation = 0

    if net:
        lbls.append('NET')

    ax.set_xticks(range(len(lbls)))
    ax.set_xticklabels(lbls, rotation=rotation)
    ax.set_ylabel(unit[0][1])

    if len(scenarios) > 1 and legend:
        handles = [mpatches.Patch(color=colors[i], label=scenarios[i]) for i in range(len(scenarios))]

        ax.legend(handles=handles, loc='best')
    ax.set_title(unit[0][0], fontsize=14)



"""
This module provides a routine that generates a list of exchange refs from a properly formatted spreadsheet
"""

from antelope import check_direction, CatalogRef, ExchangeRef
from antelope.interfaces import InvalidDirection

from .float_conv import to_float


class EmptyFlowRef(Exception):
    """
    Used to signal a blank entry
    """
    pass


class ValueIsBalance(Exception):
    """
    Used to signal when an exchange is reached that is designated as a balance.
    """
    pass


class TermDictDeprecated(Exception):
    """
    Exchanges are allowed to specify terminations, including context, compartment, defaultprovider, activitylinkid,
    or target. However, the exchanges-from-spreadsheet code is not allowed to apply any interpretation or mapping
    between the specified terminations and the targets in the foreground.  Instead, that mapping should be applied
    when the exchanges are used to create fragments.  The main reason for this is that the signature of the
    exchange interface does not support term_flow, descend, or other fragment-specific aspects.
    """
    pass


def _row_dict(sheetlike, row):
    """
    Creates a dictionary from the named row of the 'sheetlike' input.
    :param sheetlike: has row() method returns a list of strings
    :param row: int
    :return: dict k: v where k are entry.value.lower() for entry in row(0) and v are entry.value for entry in row(row)
    """
    headers = [str(k.value).lower() for k in sheetlike.row(0)]
    return {headers[i]: k.value for i, k in enumerate(sheetlike.row(row)) if i < len(headers)}


def _popanykey(dct, *keys, strict=False):
    """
    Returns the first of the listed keys to be found in the dict
    :param dct: a dict
    :param keys:
    :param strict: [False] if True, raise KeyError if we get to the end of the list and nothing was found
    :return:
    """
    for key in keys:
        if key.lower() in dct:
            k = dct.pop(key.lower())
            if k == '':
                continue
            return k
    if strict:
        raise KeyError(keys)
    return None


def _exchange_params(origin, rowdict):
    """
    This is the excel file parser that specifies how to create an exchange ref heuristically from a varied list of
    potentially valid column namings.  This could be used equally well with JSON
    :param rowdict:
    :return:
    """
    rowdict.pop('process', None)
    flow_ref = _popanykey(rowdict, 'flow', 'flowref', 'flow_ref', 'external_ref', strict=True)
    if flow_ref is None:
        raise EmptyFlowRef
    flow = CatalogRef(origin, flow_ref, entity_type='flow')
    dirn = check_direction(_popanykey(rowdict, 'direction', 'flowdir', strict=True))
    val = _popanykey(rowdict, 'value', 'amount', strict=True)
    if str(val).lower() == 'balance':
        value = ValueIsBalance
    else:
        value = to_float(val)
    unit = _popanykey(rowdict, 'unit', 'units')
    term = _popanykey(rowdict, 'target', 'defaultprovider', 'activitylinkid', 'term', 'termination', 'anchor_node')
    if term is None:
        cx = _popanykey(rowdict, 'context', 'compartment')
        if cx is not None:
            term = (cx,)  # convert to a context

    if value is ValueIsBalance:
        print('%s %s (balance) %s [%s]' % (flow, dirn, unit, term))
    else:
        print('%s %s %g %s [%s]' % (flow, dirn, value, unit, term))
    return flow, dirn, value, unit, term


def exchanges_from_spreadsheet(sheetlike, term_dict=None, node=None, origin=None):
    """
    A routine to create a list of flat exchanges (all having the same parent node) from an excel table.

    Note that this only creates a sequence of exchange refs with strings as properties, plus an inoperable process_ref
    for the parent node.  These could be used to construct a foreground model but do not by themselves constitute
    a foreground model.

    Requirements for the sheet-like object: XlrdLike / xls_tools
     - name property that returns a string
     - nrows property that reports the number of rows
     - row(index) function that returns an ordered list of entries for the specified index (0-indexed).
     - row(0) returns a header list
     - row(1) is the process's reference flow
     - subsequent rows are dependent flows

     !TODO: support multiple references with is_ref or is_reference field

    "entries" in the row() return must be "cell-like":
     - value property that returns the cell's value
     - ctype property [*not currently used*] that corresponds with xlrd.sheet.Cell:
        0=empty, 1=text, 2=number, 3=date, 4=boolean, 5=error, 6=blank
        (not clear what the difference is between empty and blank)

    Columns are interpreted as follows (case-insensitive):
    Required:
    'flow', 'flowref', 'flow_ref', 'external_ref' - in that order, used to specify flow
    'direction', 'flowdir' - in that order, used to specify direction w.r.t. parent node
    'value', 'amount' - in that order.
    This method returns unresolved catalog refs- in order for client code to use them, the flow refs must resolve (bc
    reference quantity must be known and is not required to be specified)

    Optional
    'unit', 'units' - unit of measure for the flow
    'defaultprovider', 'activitylinkid', 'target', 'term', 'termination' - anchor of an intermediate exchange
    'context', 'compartment' - context of an elementary exchange (if no anchor is found)
    if neither term nor context is found, the exchange becomes a cutoff

    Ignored:
    'process' - ignored

    Any remaining fields are passed to the exchange ref as initialization arguments.

    A 'value' whose string is lowercase-equivalent to 'balance' will add balance=True to the initialization dict and
    set the value to 0.

    :param sheetlike: an Xlrd-like object with name, nrows, and row() 
    :param term_dict: DEPRECATED.  Terminations should be handled in foreground code.
    :param node: [None] if present, use as exchange parent node for all exchanges
    :param origin: ['local.spreadsheet'] Should be provided by caller if 'node' is omitted, to give identifying 
      information to the created process
    :return:
    """
    """
    First one is the reference; subsequent ones are child flows
    sheetlike: 
    term_dict: a dictionary mapping strings found in "context", "compartment", or "flow" columns (in that order) to
     terminations.
    """
    if term_dict is not None:
        raise TermDictDeprecated('Apply term dict in foreground code when converting exchanges to fragments, not here')

    if origin is None:
        origin = 'local.spreadsheet'
    if node is None:
        proc_ref = CatalogRef(origin, sheetlike.name, entity_type='process')
    else:
        proc_ref = node

    _yield_ref_next = True

    for row in range(1, sheetlike.nrows):
        c_flow = _row_dict(sheetlike, row)
        if len(c_flow) == 0:
            continue
        try:
            flow_ref, dirn, value, units, term = _exchange_params(origin, c_flow)
        except EmptyFlowRef:
            print('==Row %02d== SKIP empty flow ref' % (row+1))
            continue
        except KeyError as e:
            print('==Row %02d== SKIP  missing one of %s' % (row+1, e.args))
            continue
        except InvalidDirection as e:
            print('==Row %02d== SKIP  invalid direction %s' % (row+1, e.args))
            continue

        if _yield_ref_next:
            # ref flow is first nonempty record
            if term is not None:
                print('Note: (%s) Reference flow cannot have specified termination: %s' % (sheetlike.name, term))
            # reference flow is unterminated
            yield ExchangeRef(proc_ref, flow_ref, dirn, value=value, unit=units, is_reference=True, **c_flow)
            _yield_ref_next = False
        else:

            if value is ValueIsBalance:
                c_flow['balance'] = True
                value = 0.0

            yield ExchangeRef(proc_ref, flow_ref, dirn, value=value, unit=units, termination=term, **c_flow)



from antelope import EntityNotFound

from .observed_mfa_study import ObservedMfaStudy

# from mfatools.aggregation.conventions import logistics_fragment_ref

from typing import Dict, Optional
from pydantic import BaseModel


def logistics_fragment_ref(fragment):
    return '%s_inbound_logistics' % fragment.external_ref


class DynamicUnitSpec(BaseModel):
    term_map: Dict[str, str] = dict()  # {flow ref: flow or terminal reference} to terminate CUTOFFS of enclosed models
    sources: Dict[str, str]  # {knob name: flow or enclosed model reference} of INFLOWS to the unit model
    sinks: Dict[str, str]  # {knob name: flow or enclosed model reference} of mass-balanced OUTFLOWS
    supplies: Dict[str, str]  # {knob name: flow or enclosed model reference} of supply use per kg of throughflow
    balance_flow: Optional[str] = None  # external ref of flow to be used for the balance output from the dynamic unit


class ParentlessKnob(Exception):
    """
    Knobs must have parents
    """
    pass


class DynamicUnitLcaStudy(ObservedMfaStudy):
    """
    This subclass takes the observed study machinery (to specify routes) adds the Dynamic Unit, which is like a synthetic MFA observation
     2-
    """
    @property
    def unit_balance_flow(self):
        return self.fg.add_or_retrieve('unit_balance', 'mass', 'Dynamic Unit Balance Flow')

    @property
    def reference_material(self):
        return self.fg.add_or_retrieve('reference_material', 'mass', 'Reference Material')

    @property
    def unit_logistics(self):
        if self.dynamic_unit:  # create if doesn't exist
            return self._fg['unit_logistics']
        else:
            raise EntityNotFound('unit_logistics')

    def make_logistics_mappings(self, logistics_mappings, add_knob=True):
        for k, v in logistics_mappings.items():
            t, stage = v
            cf = self.make_study_mapping(self.logistics_container, k, 'Input', t, stage=stage)
            if add_knob:
                # also add a logistics knob to the unit logistics model-- we could do this in general....
                knob = cf.flow.name
                knob = knob.translate(str.maketrans('/ ()', '--__'))
                print('doing unit logs %s' % knob)
                self._add_unit_knob(knob, cf.flow, 'Input', self.unit_logistics, descend=False)

    @property
    def dynamic_unit(self):
        """
        The Dynamic Unit is a structure that enables scenarios to be created within the Mfa study superstructure
        that simulate unit activities in the broader structure.  The unit is installed as a study object within the
        activity container.  It has the name 'unit_model' and the following structure:

        <--[unit ref]--O  {'unit_model'} (activity)
                       |
                       +-=>=-O reference material (mass balance)
                             |
                             +-=>=-O reference material (mass balance) {'unit-dynamic_sinks'}
                             |     |
                             |     +-=>=-  dynamic balance (mass balance) (cutoff)
                             |
                             +-[1.0]>-O unit ref {'unit-dynamic_supplies'}
                                      |
                                      +-[1.0]>-O reference material {'unit_logistics'}

        Upon creation, the dynamic unit will not do anything because its only child flow is a mass balance.
        The user makes the dynamic useful by adding "knobs" to turn in the following places:

        add_unit_source-- adds an Input child flow to 'unit_model', to drive the mass balance
        add_unit_sink-- adds an output child flow to 'unit-dynamic_sinks' to drive a downstream process
         (note, dynamic sinks are driven by the balance of sources)
        add_unit_supply-- adds an input child flow to 'unit-dynamic_supplies'
         (note, dynamic supplies are with respect to a unit magnitude of the reference material)
        add_logistics_route-- adds an input child flow to 'unit_logistics'
         (note, logistics are with respect to a unit magnitude of the reference material)

        After the knobs have been created, the user can specify knob "settings" for each scenario.

        :return:
        """
        try:
            return self._fg.get('unit_model')
        except EntityNotFound:
            unit_ref = self.new_activity_flow('Unit Reference Flow', external_ref='unit_reference_flow')
            unit = self._fg.new_fragment(unit_ref, 'Output', name='%s - Dynamic Unit' % self._ref, external_ref='unit_model')
            b = self._fg.new_fragment(self.reference_material, 'Output', parent=unit, balance=True)
            c = self._fg.new_fragment(self.reference_material, 'Output', parent=b, balance=True, external_ref='unit-dynamic_sinks')
            self._fg.new_fragment(self.unit_balance_flow, 'Output', parent=c, balance=True)
            d = self._fg.new_fragment(unit_ref, 'Output', parent=b, exchange_value=1.0, external_ref='unit-dynamic_supplies')
            self._fg.observe(d)  # lock in the 1.0 exchange value
            # d.to_foreground()  # this is now accomplished in new fragment constructor via set_parent()
            rl = self._fg.new_fragment(self.reference_material, 'Output', parent=d, exchange_value=1.0, external_ref='unit_logistics')
            # rl.terminate(NullContext)  # replaces to_foreground() we do not want this going to context!
            self._fg.observe(rl)  # lock in the 1.0 exchange value

            self._fg.observe(self.activity_container, termination=unit, scenario='Unit')
        return self._fg.get('unit_model')

    def add_unit_source(self, knob, source, descend=False, term_map=None):
        """
        Simply installs the named flow as a source knob-- an inflow to the unit model.  If the supplied source
        is a flow, it is assumed to be terminated in the study layer.  If it is a fragment, then its inventory
        flows are terminated according to the term_map.

        :param knob: a string
        :param source:
        :param descend: [False] whether the traversal should descend [True] or aggregate [False] the knob
        :param term_map: if source is a model, mapping of source's cutoffs to background terminations or flows
        :return:
        """
        ''' The following comment appears to be false and was removed:
        Note: term_map entries must be ENTITIES and not REFs.  This is b/c of the mfa vs models vs study conundrum.
        '''
        return self._add_unit_knob(knob, source, 'Input', self.dynamic_unit, descend, term_map)

    def add_unit_sink(self, knob, sink, descend=False, term_map=None):
        """

        :param knob:
        :param sink:
        :param descend: [False] whether the traversal should descend [True] or aggregate [False] the knob
        :param term_map:
        :return:
        """
        parent = self._fg.get('unit-dynamic_sinks')
        return self._add_unit_knob(knob, sink, 'Output', parent, descend, term_map)

    def add_unit_supply(self, knob, supply, direction='Input', descend=False, term_map=None):
        """
        In the current construction, supply knobs MUST be reported PER kg of flow through the dynamic unit.
        
        (alternative design would be to make the dynamic supply a child of the dynamic unit, and they would be 
        absolute amounts per unit.) 
        
        :param knob:
        :param supply:
        :param direction: default Input
        :param descend: [False] whether the traversal should descend [True] or aggregate [False] the knob
        :param term_map:
        :return:
        """
        parent = self._fg.get('unit-dynamic_supplies')
        return self._add_unit_knob(knob, supply, direction, parent, descend, term_map)

    '''
    def add_logistics_route(self, flow, provider, descend=False, term_map=None, **kwargs):
        c = super(DynamicUnitLcaStudy, self).add_logistics_route(flow, provider, descend=descend, **kwargs)
        knob = c.flow.name
        knob = knob.replace('/', '_')
        return self._add_unit_knob(knob, c.flow, 'Input', self.unit_logistics, descend, term_map=term_map)
    '''

    def _add_unit_knob(self, knob, entry, direction, parent, descend=None, term_map=None):
        """

        :param knob: a string knob name-- the "knob" fragment will be assigned this external_ref
        :param entry: what does the knob twiddle? string gets resolved to entity.
         flow- will just be a cutoff, to be plumbed to the activity/logistics/product layers
         fragment- will get terminated
        :param direction: sources should be 'Input', sinks should be 'Output'
        :param parent: where does the knob get added?
        :param descend: whether the traversal should descend [True] or aggregate [False] the knob. (term_map flows are
         always non-descend)
        :param term_map:
        :return:
        """
        if parent is None:
            raise ParentlessKnob(knob)
        try:
            k = self._fg.get(knob)
        except EntityNotFound:
            resolved_entry = self._resolve_term(entry)
            if resolved_entry.entity_type == 'flow':
                k = self._fg.new_fragment(resolved_entry, direction, parent=parent, value=0, external_ref=knob)
            elif resolved_entry.entity_type == 'fragment':
                k = self._fg.new_fragment(resolved_entry.flow, resolved_entry.direction, parent=parent, value=0,
                                          external_ref=knob)
                k.terminate(resolved_entry, descend=descend)
                if term_map:
                    self._add_child_flows(k, resolved_entry, term_map)
            else:
                raise TypeError('Improper type %s (%s)' % (type(resolved_entry), resolved_entry))
        return k

    def _add_child_flows(self, frag, term, dynamic_outputs):
        for k in term.cutoffs():
            if not k.is_reference:
                if k.flow.external_ref in dynamic_outputs:
                    v = self._resolve_term(dynamic_outputs[k.flow.external_ref])
                    if v.entity_type == 'flow':
                        o = self._fg.new_fragment(k.flow, k.direction, parent=frag)
                        self._fg.new_fragment(v, k.direction, parent=o, balance=True)
                    elif v.entity_type == 'fragment':
                        o = self._fg.new_fragment(k.flow, k.direction, parent=frag)
                        o.terminate(v, term_flow=v.flow, descend=False)

    def set_unit_balance(self, flow):
        bf = self._resolve_term(flow)
        cf = self.fg['unit-dynamic_sinks'].balance_flow
        df = cf.balance_flow
        if df is None:
            self.fg.new_fragment(bf, 'Output', parent=cf, balance=True)
        else:
            df.flow = bf

    def install_observation_model(self, prov_frag, scope=None):
        super(DynamicUnitLcaStudy, self).install_observation_model(prov_frag, scope=scope)

        log_ref = logistics_fragment_ref(prov_frag)
        prov_log = self.data[log_ref]  # should also be a convention?
        if prov_log is None:
            print('No provincial logistics found!')
        else:
            self.unit_logistics.clear_termination('Unit-%s' % scope)
            self.unit_logistics.terminate(prov_log, 'Unit-%s' % scope)

    def make_dynamic_unit(self, unit_spec: DynamicUnitSpec, descend=False):
        for k, v in unit_spec.sources.items():
            self.add_unit_source(k, v, descend=descend, term_map=unit_spec.term_map)
        for k, v in unit_spec.sinks.items():
            self.add_unit_sink(k, v, descend=descend, term_map=unit_spec.term_map)
        for k, v in unit_spec.supplies.items():
            self.add_unit_supply(k, v, descend=descend, term_map=unit_spec.term_map)
        if unit_spec.balance_flow:
            self.set_unit_balance(unit_spec.balance_flow)






from collections import defaultdict
from pandas import DataFrame, MultiIndex


def _mag(ff):
    if ff.fragment.direction == 'Input':
        return ff.magnitude
    return -1.0 * ff.magnitude


class FragmentViewer(object):
    """
    Show the exterior flows coming out of a fragment
    """
    def _generate(self):
        tl = defaultdict(list)
        # sort fragment flows by flow
        for ff in self._frag.traverse(scenario=self._scen):
            if not ff.term.is_null:
                continue
            tl[ff.fragment.flow].append(ff)

        ks = list(tl.keys())  # list of flows
        self._mi = MultiIndex.from_tuples((k.name, k.unit) for k in ks)
        self._df = DataFrame(({k.fragment.reference_entity: _mag(k) for k in tl[z]} for z in ks), index=self._mi)
        self._mc = MultiIndex.from_tuples((k['Name'], k['StageName']) for k in self._df.columns)
        self._df.columns = self._mc
        self._df.sort_index(axis=0, level=1, inplace=True)
        self._df.sort_index(axis=1, level=1, inplace=True)

    def __init__(self, fragment, scenario=None, total=False):
        self._frag = fragment
        self._scen = scenario

        self._df = None  # dataframe
        self._mc = None  # multicolumn
        self._mi = None  # multiindex

        self._generate()

        if total:
            self.add_total()

    @property
    def df(self):
        return self._df

    @property
    def index(self):
        return self._mi

    @property
    def columns(self):
        return self._mc

    def stages(self):
        return sorted(set(self._mc.get_level_values(1)))

    def stage_table(self, stage, dropna='all', total=True):
        dd = self._df.loc[:, self._df.columns.get_level_values(1) == stage].dropna(how=dropna)
        if total:
            dd['Total'] = dd.sum(axis=1, skipna=True)
        return dd

    def add_total(self):
        self._df['Total'] = self._df.sum(axis=1, skipna=True)



"""
A function to read a list of observations from a spreadsheet and apply them to models found in a foreground.

The spreadsheet should have the following fields:
Specifiers:
 activity - external ref of activity
 child_flow - external ref of flow-- all child flows with the named flow will be affected.  if omitted, apply to the
  reference flow
 scenario - scenario name under which the observation should be applied

Arguments:
 anchor - new anchor. 'cutoff' or 'CUTOFF' to cutoff flow
 descend - for anchor. Should be case-insensitive true or false or 0 or 1. blank = None = unspecified.
 parameter - new exchange value
 units - units for exchange value

"activity" must refer to a fragment, and is required.
If "child_flow" is omitted, the observation will be applied to the named activity.
If "scenario" is omitted, the observation will be applied to the default scenario.

any of the arguments may be included; if no arguments are included, the row will have no effect.
special semantics:
If "child_flow" is "*" and the "descend" specification is not None, then the specification is applied to all child flows

"""
from .float_conv import to_float


def _cutoff(anc):
    if isinstance(anc, str):
        if anc.lower() == 'cutoff':
            return True
    return False


def _descend(desc):
    if desc is None:
        return None
    elif isinstance(desc, str):
        if desc.lower() in ('false', '0'):
            return False
        elif desc.lower() in ('true', '1'):
            return True
        elif desc.lower() == 'none':
            return None
    return bool(desc)


class ObservationsFromSpreadsheet(object):
    def _errmesg(self, ssr, message):
        if self.quiet:
            return
        print('OBS: Row %d: %s' % (ssr, message))

    def _mesg(self, obj, sc, message):
        if self.quiet:
            return
        print('OBS: %5.5s {%s} %s' % (obj.uuid, sc, message))

    def __init__(self, fg, sheet, quiet=False):
        """

        :param fg: a foreground
        :param sheet: an Xlrd-like.
        :param quiet: default False
        :return:
        """
        self.fg = fg
        self.sheet = sheet
        self.quiet = quiet

    def _handle_anchor(self, ssr, obj, row):
        # anchor
        sc = row.get('scenario')

        anc = row.get('anchor')

        if anc:
            if _cutoff(anc):
                for o in obj:
                    o.clear_termination(scenario=sc)
                    self._mesg(o, sc, 'to cutoff')
                    return
            desc = _descend(row.get('descend'))
            a_o = row.get('anchor_origin')
            a_f = row.get('anchor_flow')
            if a_o:
                anchor = self.fg.catalog_ref(a_o, anc)
            else:
                anchor = self.fg[anc]
                if anchor is None:
                    self._errmesg(ssr, 'Anchor %s not found' % anc)
                    return
            for o in obj:
                o.clear_termination(scenario=sc)
                if desc is None:
                    desc = o.term.descend
                self.fg.observe(o, scenario=sc, anchor_node=anchor, anchor_flow=a_f, descend=desc)
                mesg = 'anchor to %s' % anc
                if a_f:
                    mesg += ' (anchor flow %s)' % a_f
                if desc is not None:
                    mesg += ' (descend %s)' % bool(desc)
                self._mesg(o, sc, mesg)
        else:
            desc = _descend(row.get('descend'))

            if desc is not None:
                for o in obj:
                    if o.termination(sc) is o.term:
                        o.terminate(o.term.term_node, scenario=sc, term_flow=o.term.term_flow, descend=desc)
                    else:
                        o.termination(sc).descend = desc
                    self._mesg(o, sc, 'descend to %s' % desc)

    def _handle_ev(self, obj, row):
        sc = row.get('scenario')
        ev = row.get('parameter')
        units = row.get('units')
        if ev is not None:
            ev = to_float(ev)
            for o in obj:
                self.fg.observe(o, scenario=sc, exchange_value=ev, units=units)
                mesg = 'observing %g' % ev
                if units is not None:
                    mesg += ' %s' % units
                self._mesg(o, sc, mesg)

    def apply(self):
        for i in range(1, self.sheet.nrows):
            ssr = i + 1
            row = self.sheet.row_dict(i)
            if row.get('activity') is None:
                self._errmesg(ssr, 'Skipping blank row')
                continue
            act = self.fg[row['activity']]
            if act is None:
                self._errmesg(ssr, 'Activity %s not found' % row['activity'])
                continue
            if row.get('child_flow'):
                if row['child_flow'] == '*':
                    obj = [cf for cf in act.child_flows]
                    p = row.pop('parameter', None)
                    a = row.pop('anchor', None)
                    if p or a:
                        self._errmesg(ssr, 'ignoring specs for descend special "*"')
                else:
                    cf = self.fg[row['child_flow']]
                    if cf is None:
                        self._errmesg(ssr, 'Child flow %s not found' % row['child_flow'])
                        continue
                    if cf.entity_type == 'flow':
                        obj = list(act.children_with_flow(cf, recurse=True))
                    else:
                        obj = [cf]
            else:
                obj = [act]
            self._handle_anchor(ssr, obj, row)
            self._handle_ev(obj, row)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return



from .lca_model_runner import LcaModelRunner


class SimpleModelBuilder(LcaModelRunner):
    """
    We want to take in process refs, build fragments out of them, and expand them one level deep.
    """
    def __init__(self, fg, **kwargs):
        super(SimpleModelBuilder, self).__init__(**kwargs)

        self._fg = fg

        self._frags = dict()

    def add_case(self, name, p_ref):
        if name in self._scenarios:
            raise KeyError('Name already exists')
        frag = self._fg.create_process_model(p_ref)
        self._fg.extend_process(frag, multi_flow=True)
        self.add_scenario(name)
        self._frags[name] = frag

    def _run_scenario_lcia(self, scenario, lcia, **kwargs):
        frag = self._frags[scenario]
        return frag.fragment_lcia(lcia, **kwargs)


"""
The purpose of an upgrade manager is to migrate nodes in a foreground ('kit') from one origin to another.
We can imagine a number of different use cases:

 - USLCI: same process, different origin
 - ecoinvent: same flow, with geographically differentiated targets
 - ecoinvent-TO-other: requires a mapping service (this is actually qdb, I think)

for now we will use subclassing
"""
from antelope import enum, MultipleReferences, NoReference  # is this *always* an interactive tool?
from antelope_foreground.models import Anchor


class NoCandidates(Exception):
    pass


class TooManyCandidates(Exception):
    pass


class UpgradeManager(object):
    """
    foreground query
    source origin
    target query

    """
    _strategy = 'match_name_and_spatial_scope'

    _strategies = ('match_name_and_spatial_scope',
                   'match_name',
                   'same_id',
                   'targets',
                   'node_targets')

    def __init__(self, fg, origin, query, scenario=None, to_scenario=None, strategy=None, autoskip=False):
        """

        :param fg: foreground query that contains the nodes to be upgraded
        :param origin:
        :param query:
        :param scenario: [None] scenario to revise; "from" scenario
        :param to_scenario: scenario name to use during observation. 'None' will not be accepted if scenario is also
        None-- specify 'default' if you want to overwrite the default scenario
        :param strategy:
        :param autoskp: [False] If True, skip entries that have already been observed under to_scenario
        """
        self._fg = fg
        self._src = origin
        self._q = query
        self._scenario = scenario
        self._to_scenario = to_scenario
        self._autoskip = autoskip

        self._nodes = tuple(n for n in self._fg.nodes(origin=self._src) if n.scenario == scenario)
        self._pending = list(range(len(self._nodes)))
        self._working = None  # stores the INDEX into the node being reviewed
        self._rxs = ()  # chosen candidate's rx
        self._rx = None  # stores the selected termination
        self._candidates = None
        self._exception = None
        self._completed = [False] * len(self._nodes)
        self._error = [None] * len(self._nodes)

        self.strategy = strategy

        self.next()

    @property
    def strategy(self):
        return self._strategy

    @strategy.setter
    def strategy(self, value):
        if value is not None:
            if value in self._strategies:
                print('setting strategy "%s"' % value)
                self._strategy = value

    @property
    def nodes(self):
        for n in self._nodes:
            yield n

    def nodes_for(self, fragment_ref):
        for n in self._nodes:
            if n.node.entity_id == fragment_ref:
                yield n

    def is_pending(self, x):
        return x in self._pending

    def is_completed(self, x):
        return bool(self._completed[x])

    def error(self, x):
        return self._error[x]

    def next(self, x=None):
        if self._working and self._exception:
            self._error[self._working] = self._exception
        self._working = None

        self._candidates = self._exception = self._rx = self._rxs = None
        self._working = self._find_next(x)
        if self._working is None:
            print('Queue has been exhausted')
            return
        print('Next up: %d: %s -# %s' % (self._working, self.current.node.name, self.current.anchor.name))

    def _find_next(self, x=None):
        rtn = None
        while rtn is None:
            if x is None:
                try:
                    rtn = self._pending[0]
                except IndexError:
                    return
            else:
                if x not in self._pending:
                    raise IndexError
                rtn = x
                x = None

            if self._to_scenario is not None:
                if self._to_scenario in self._nodes[rtn].node.scenarios(recurse=False):
                    term = self._nodes[rtn].node.termination(self._to_scenario)
                    if self._autoskip:
                        print('%d: Skipping observed entry (%s: %s)' % (rtn, self._to_scenario, term.term_node))
                        self._pending.remove(rtn)
                        rtn = None
                    else:
                        print('Current [%s] already observed as: %s' % (self._to_scenario,
                                                                        term.term_node))
        return rtn

    @property
    def rx(self):
        if self._rx is None:
            self.pick()
        return self._rx

    @property
    def completed(self):
        for i, n in enumerate(self._nodes):
            if self._completed[i]:
                yield n

    @property
    def errored(self):
        for i, n in enumerate(self._nodes):
            if bool(self._error[i]):
                yield n

    @property
    def current(self):
        if self._working is not None:
            return self._nodes[self._working]

    def candidates(self):
        if self._candidates:
            enum(self._candidates)

    """
    different upgrade strategies
    """
    def match_name(self):
        """
        Match the name of the current anchor
        :return:
        """
        return self._q.processes(name='^%s$' % self.current.anchor.term_node['name'])

    def match_name_and_spatial_scope(self):
        """
        Match the name and spatial scope of the current anchor
        :return:
        """
        return self._q.processes(name='^%s$' % self.current.anchor.term_node['name'],
                                 spatialscope='^%s$' % self.current.anchor.term_node['spatialscope'])

    def same_id(self):
        """
        Retrieve the process with the same external_ref from a different query
        :return:
        """
        return [self._q.get(self.current.anchor.term_node.external_ref)]

    def targets(self):
        """
        Retrieve targets that match the current anchor (flow and direction)
        :return:
        """
        return self._q.targets(self.current.anchor.term_flow, direction=self.current.anchor.direction)

    def node_targets(self):
        """
        Retrieve targets that match the current node (flow, disregarding direction)
        :return:
        """
        return self._q.targets(self.current.node.flow)

    def _run_attempt(self, strategy):
        """
        :return:
        """
        if strategy is None:
            strategy = self.strategy
        self._candidates = enum(getattr(self, strategy)())

    def attempt(self, x=None, strategy=None):
        if x:
            self.next(x)
        try:
            self._run_attempt(strategy)  # populates _candidates
            if len(self._candidates) > 0:
                return True
            print('NoCandidates')
            self._exception = NoCandidates()
            return False
        except Exception as e:
            self._exception = e
            print(e)
            return False

    def rxs(self, n=0):
        return self._candidates[n].references()

    def pick(self, n=0, ref_flow=None):
        """
        subclass: process with ambiguous ref
        :param n:
        :param ref_flow:
        :return:
        """
        self._rxs = tuple(self._candidates[n].references())
        try:
            self._rx = self._candidates[n].reference(ref_flow)  # if ref flow is None and len(rx) is 1, will return it
        except MultipleReferences as m:
            if ref_flow is None:
                try:
                    self._rx = self._candidates[n].reference(self.current.anchor.term_flow)
                    return
                except NoReference:
                    pass
            raise m
        except NoReference:
            raise

    def pick_rx(self, k=0):
        if self._rxs is None:
            self.pick()
        self._rx = self._rxs[k]

    def observe_current_node(self, to_scenario=None, descend=None, **kwargs):
        """
        The specified candidate is chosen. moves the identified working node from pending into completed.
        :param to_scenario: if None, self._scenario is used for the observation. If self._scenario is None, a scenario must
        be provided. Use 'default' if the observation is intended to reset the default anchor (i.e. scenario=None)
        :param descend: specified in anchor. if None, the branch's current anchor is used
        :param kwargs: passed to fg.observe()
        :return:
        """
        if self._candidates is None:
            if self.attempt() is False:
                raise self._exception
        if self._rx is None:
            self.pick()
        if descend is None:
            descend = self.current.anchor.descend
        if to_scenario is None:
            if self._scenario is None and self._to_scenario is None:
                raise ValueError("supply 'default' if you want to rewrite the default scenario")
            to_scenario = self._to_scenario

        try:
            print("observing %s" % self.rx)
            self._fg.observe(self.current.node, scenario=to_scenario,
                             anchor_node=self.rx.process, anchor_flow=self.rx.flow, descend=descend, **kwargs)
            self.finish()

        except Exception as e:
            self._exception = e
            print(e)
            return False

        return True

    def finish(self):
        self._completed[self._working] = True
        self._pending.remove(self._working)
        self.next()

    def skip(self):
        """
        Remove from pending list without marking complete
        :return:
        """
        self._pending.remove(self._working)
        self.next()

    def fail(self):
        """
        remove from
        :return:
        """
        self._completed[self._working] = False
        self._pending.remove(self._working)
        self.next()

    def run(self, to_scenario=None, descend=None, **kwargs):
        while len(self._pending) > 0:
            if self.attempt():
                if len(self._candidates) > 1:
                    self._exception = TooManyCandidates()
                    self.fail()
                else:
                    if self.observe_current_node(to_scenario=to_scenario, descend=descend, **kwargs):
                        pass
                    else:
                        self.fail()
            else:
                self.fail()
        print('%d success\n%d fail' % (sum(self._completed), len(list(self.errored))))

    def reset_pending(self):
        self._pending = [k for k in range(len(self._nodes)) if not self.is_completed(k)]



from collections import namedtuple
from antelope import EntityNotFound, UnknownOrigin


ScenarioParam = namedtuple('ScenarioParam', ('fragment', 'scenario', 'value'))  # need uncertainty params.. eventually
ScenarioTermination = namedtuple('ScenarioTermination', ('fragment', 'scenario', 'termination', 'term_flow', 'descend'))



class XlsxScenarioUpdater(object):
    """
    Apply scenarios to the fragments found in an foreground archive.
    Operates as a context manager:
    >>> with XlsxScenarioUpdater(fg, xlsx, *sheets) as a:
    >>>    a.apply()
    NOTE: all pre-existing scenarios in fg are removed and replaced with the ones found in *sheets

    There are two different patterns for a sheet listed in *sheets (both can coexist) (all case-insensitive):
     - observations: must have columns 'fragment', 'scenario', 'observedValue'
     - terminations: must have columns 'fragment', 'scenario', 'termination', 'term_flow'
       term_flow can be omitted but is obligatory in many cases
    """

    @staticmethod
    def _grab_value(cell):
        value = cell.value
        if isinstance(value, str):
            if value == '':
                value = None
        return value

    def _read_scenarios(self, sheet):
        if sheet in self._xl.sheet_names():
            sh = self._xl.sheet_by_name(sheet)
            headers = [k.value.lower() for k in sh.row(0)]
            _ev = 'observedvalue' in headers
            _tm = 'termination' in headers
            for i in range(1, sh.nrows):
                row = [self._grab_value(k) for k in sh.row(i)]
                d = {headers[k]: row[k] for k in range(min([len(row), len(headers)]))}
                if 'fragment' not in d or d['fragment'] is None:
                    continue
                if _ev:
                    if d.get('observedvalue') is not None:
                        self._params.append(ScenarioParam(d['fragment'], d['scenario'], float(d['observedvalue'])))
                if _tm:
                    if d.get('termination') is not None:
                        desc = {'true': True,
                                'false': False,
                                '0': None}[d.get('descend', '0').lower()]  # map to bool

                        self._terms.append(ScenarioTermination(d['fragment'], d['scenario'], d['termination'],
                                                               d.get('term_flow'), desc))

    def _print(self, *args):
        if not self._quiet:
            print(*args)

    def __init__(self, fg, xlrd_like, *scenario_sheets, terminations=True, quiet=True):

        self._fg = fg
        self._xl = xlrd_like

        self._params = []
        self._terms = []

        self._do_term = terminations
        self._quiet = quiet

        self._model_data = []
        if len(scenario_sheets) == 0:
            scenario_sheets = ['scenarios-observations', 'scenarios-terminations']

        for sheet in scenario_sheets:
            self._read_scenarios(sheet)
        self._unrec = []
        self._print('Loaded %d scenario params and %d scenario terminations' % (len(self._params), len(self._terms)))

    def _apply_params(self):
        for param in self._params:
            frag = self._fg[param.fragment]
            if frag is None:
                self._print('Skipping %s' % param.fragment)
                continue
            scen = param.scenario
            if isinstance(scen, str) and scen.lower() == 'none':
                scen = None
            self._fg.observe(frag, scenario=scen, exchange_value=param.value)
            self._print('%s: %g [%s]' % (frag.name, param.value, scen))

    def _apply_terms(self):
        for term in self._terms:
            frag = self._fg[term.fragment]
            scen = term.scenario
            try:
                t = self._fg.get(term.termination)
            except (EntityNotFound, UnknownOrigin):
                self._unrec.append((term.fragment, term.termination))
                continue
            if scen.lower() == 'none':
                scen = None
                print('warning: changing default termination for %s' % frag)
                frag.clear_termination(None)

            tf = term.term_flow
            if tf is not None:
                tf = self._fg.get(tf)
            frag.terminate(t, scenario=scen, term_flow=tf, descend=term.descend)
            self._print('%s %s %s (%s) [%s]' % (frag.name, frag.termination(scen), t.name, tf, scen))

    def apply(self):
        self._fg.clear_scenarios(terminations=self._do_term)
        self._apply_params()
        if self._do_term:
            self._apply_terms()

        if len(self._unrec) > 0:
            print('Unrecognized Terminations: ')
            for frag, term in self._unrec:
                print('  [%s] -> %s' % (frag, term))

    def __enter__(self):
        """Return self object to use with "with" statement."""
        return self

    def __exit__(self, *args):
        if hasattr(self._xl, 'release_resources'):
            self._xl.release_resources()
        self._xl = None



from antelope import EntityNotFound
from xlstools.google_sheet_reader import GoogleSheetReader
from googleapiclient.errors import HttpError

INDEX_HEADINGS = ('Abbreviation', 'Name', 'ShortName', 'Method', 'Category', 'Indicator', 'unit', 'uuid', 'Comment', 'Notes')
FACTOR_HEADINGS = ('flowable', 'context', 'ref_quantity', 'ref_unit', 'locale', 'value')


class QdbGSheetClient(object):
    """
    This class allows us to do the following things:
     - read in a quantity definition from a google sheet and create a quantity in a designated foreground (?)
     - assign characterization factors from the spreadsheet to the quantity
     - assign characterization factors from the quantity to the spreadsheet
     - update quantity definition in spreadsheet
     - create a new quantity and define it in both the spreadsheet and the foreground

    Other things that could be done:
     - given a list of flow specs, return LCIA results
     - term manager stuff

    The reference Qdb could be entirely constituted from LCIA specifications and synonym specifications

    The essential challenge of this class is in establishing correspondence between two different information stores:
     - the local contents of the foreground
     - the static contents of the spreadsheet.

    The information can be synchronized in either direction.  Any time a synch is performed, the external_ref of the
    local quantity is stored as a mapping to the abbreviation of the remote quantity.

    The external_ref used for a quantity in a local foreground can vary.  Any time
    """
    def __init__(self, fg, sheet_id, json_credential=None, credential_file=None, **kwargs):
        self._fg = fg
        cred = json_credential or credential_file
        self._xls = GoogleSheetReader(cred, sheet_id)
        self._args = kwargs
        self._qsheet = None
        self._qs = []  # list of indicators in sequence
        self._qd = dict()  # dict of ref to indicator
        self._ext_ref_mapping = dict()  # custom mapping of fg external ref to indicator name "item"
        self.load_quantities()

    @property
    def fg(self):
        return self._fg

    @property
    def static(self):
        return False

    def load_quantities(self):
        if 'Quantities' not in self._xls.sheet_names():
            self._xls.create_sheet('Quantities')
            self._xls.write_row('Quantities', 0, INDEX_HEADINGS)
        self._qsheet = self._xls.sheet_by_name('Quantities')
        assert tuple(k.value for k in self._qsheet.row(0)) == INDEX_HEADINGS, "columns do not match specification!"
        qs = [self._qsheet.row_dict(i) for i in range(1, self._qsheet.nrows)]
        self._qd = {k['Abbreviation']: k for k in qs}
        self._qs = [k['Abbreviation'] for k in qs]

    @property
    def quantities(self):
        for q in self._qs:
            yield self._qd[q]

    def __getitem__(self, item):
        """
        Returns *remote metadata* for the quantities known to the spreadsheet
        :param item:
        :return:
        """
        try:
            return self._qd[item]
        except KeyError:
            return self._qd[self._ext_ref_mapping[item]]

    def _update_mapping(self, item, external_ref=None):
        if external_ref is None:
            if item in self._ext_ref_mapping:
                # user specified an external ref
                external_ref = item
                item = self._ext_ref_mapping[external_ref]
            else:
                # user specified an abbreviation without a mapping
                if item not in self._qd:
                    raise KeyError('Unknown item %s' % item)
                try:
                    external_ref = next(k for k, v in self._ext_ref_mapping.items() if v == item)
                except StopIteration:
                    # the item is un-mapped; establish an epoynmous mapping
                    external_ref = item
                    self._ext_ref_mapping[external_ref] = item
        else:
            self._ext_ref_mapping[external_ref] = item
        return item, external_ref

    def create_or_update_quantity(self, item, name=None, ref_unit=None, external_ref=None, **kwargs):
        item, external_ref = self._update_mapping(item, external_ref)
        if item in self._qd:
            dat = self._qd[item]
        else:
            if name is None or ref_unit is None:
                raise ValueError('Must provide name and ref unit to create new method')
            dat = {'Abbreviation': item,
                   'unit': ref_unit}
            self._qd[item] = dat
            self._qs.append(item)
        if name is not None:
            dat['Name'] = name

        for n in INDEX_HEADINGS:
            if n in kwargs:
                dat[n] = kwargs[n]

        ent = self.fetch_quantity(item, external_ref)
        self.write_quantity_metadata(item, external_ref)
        return ent

    def fetch_quantity(self, item, external_ref=None):
        """
        Updates the quantity in the local foreground according to the specification on the google sheet.
        :param item:
        :param external_ref:
        :return:
        """
        item, external_ref = self._update_mapping(item, external_ref)
        dat = {**self._qd[item]}
        if self.fg[external_ref] is None:
            ent = self.fg.new_quantity(dat.pop('Name'), ref_unit=dat.pop('unit'),
                                       external_ref=external_ref, synonyms=(item,), **dat)
        else:
            ent = self.fg[external_ref]
            for k, v in dat.items():
                if k == 'unit':
                    if ent.unit != v:
                        print('%s: Cannot update reference unit for existing quantity (%s)' % (ent, v))
                else:
                    ent[k] = v
        return ent

    def write_quantity_metadata(self, item, external_ref=None):
        """
        Tricky.  Do we specify the local quantity, or the remote abbreviation?  easy on reflection: the abbreviation
        is required so go ahead and make it the first positional parameter. This has the advantage of being consistent
        with the whole rest of the API (except that _update_mapping implicitly accepts external_ref)
        :param item:
        :param external_ref:
        :return:
        """
        item, external_ref = self._update_mapping(item, external_ref)
        ent = self.fg[external_ref]
        dat = {'Abbreviation': item}
        for k in INDEX_HEADINGS:
            if k == 'Abbreviation':
                continue
            elif k == 'unit':
                dat[k] = ent.unit
            else:
                dat[k] = ent.get(k)
        if item not in self._qd:
            self._qs.append(item)
            self._qd[item] = dat
        row = self._qs.index(item) + 1
        row_data = [dat[k] for k in INDEX_HEADINGS]
        self._xls.write_row('Quantities', row, row_data)

    def _factor_headings(self, item):
        indicator = self._qd[item]['Indicator']
        columns = FACTOR_HEADINGS + (indicator,)
        return columns

    def _create_or_retrieve_cf_sheet(self, item):
        columns = self._factor_headings(item)
        if item not in self._xls.sheet_names():
            self._xls.create_sheet(item)
        try:
            self._xls.write_row(item, 0, columns)
        except HttpError:
            print('No write access; not updating column headings for %s' % item)
        return self._xls.sheet_by_name(item)

    def update_cfs(self, item, external_ref=None):
        """
        Appply static gsheet CFs to local quantity

        :param item:
        :param external_ref: reference for local quantity
        :return:
        """
        item, external_ref = self._update_mapping(item, external_ref)
        ent = self.fetch_quantity(item, external_ref)
        sheet = self._create_or_retrieve_cf_sheet(item)
        for i in range(1, sheet.nrows):
            cf = sheet.row_dict(i)
            try:
                rq = self.fg.get_canonical(cf['ref_quantity'])
            except EntityNotFound:
                print('Ref quantity %s not found' % cf['ref_quantity'])
                continue
            value = cf['value'] * rq.convert(to=cf['ref_unit'])  # check this!  if the CF is 45 points per gram and
            # the ref unit is kg, then that's 45,000 points per kg
            ent.characterize(cf['flowable'], rq, value, context=cf['context'], location=cf['locale'],  # why "location"?
                             overwrite=True)

        return ent

    def write_quantity_cfs(self, item, external_ref=None):
        """
        Write local CFs to the google sheet (destructively-- run update_cfs first to not lose data)
        :param item:
        :param external_ref:
        :return:
        """
        item, external_ref = self._update_mapping(item, external_ref)
        ent = self.fetch_quantity(item, external_ref)

        self._create_or_retrieve_cf_sheet(item)

        data = []
        for cf in sorted(ent.factors(), key=lambda x: (x.flowable, x.context.name)):
            for locale in cf.locations:
                row = (cf.flowable, cf.context.name, cf.ref_quantity['name'], cf.ref_quantity.unit,
                       locale, cf[locale], ent.unit)
                data.append(row)
        self._xls.clear_region(item, start_row=1)
        self._xls.write_rectangle_by_rows(item, data, start_row=1)



import pydot
from antelope import check_direction
from antelope_core.autorange import AutoRange
from math import log10


def line_break(f, start=40):
    t = ''
    while f.find(' ', start) >= start:
        x = f.find(' ', start)
        t += f[:x]
        t += '\n'
        f = f[x+1:]
    t += f
    return t


class ModelGraph(object):
    """
    Creates a model graph in graphviz dot format using pydot.
    """
    def _flowtext(self, text):
        return line_break(text, start=self._text_width)

    def _dir(self, frag):
        """
        Returns the direction node
        :param frag:
        :return:
        """
        if frag.is_balance:
            _d = pydot.Node(frag.uuid[:10] + '_d', shape="box", style="dotted", label="balance")
        else:
            shape = {'Output': 'rarrow',
                     'Input': 'larrow'}[check_direction(frag.direction)]
            if bool(self.values):
                ev = frag.exchange_value(self._scenario or True)
                if bool(self.autoscale) and frag.flow.unit != 'Item(s)':
                    au = AutoRange(abs(ev), bias=-0.01)
                    label = '%.3g %s' % (au.adjust(ev), au.adj_unit(frag.flow.unit))
                else:
                    if ev == 0:
                        m = 2  # we just want a '0.0'
                    else:
                        m = log10(abs(ev))
                    if abs(m) <= 3:
                        if m > 1:
                            label = '%.1f %s' % (ev, frag.flow.unit)
                        elif m > -1:
                            label = '%.2f %s' % (ev, frag.flow.unit)
                        else:
                            label = '%.3f %s' % (ev, frag.flow.unit)
                    else:
                        label = '%.3g %s' % (ev, frag.flow.unit)
            else:
                label = frag.flow.unit
            _d = pydot.Node(frag.uuid[:10] + '_f', shape=shape, label=label)
        self._graph.add_node(_d)
        return _d

    def __init__(self, model, scenario=None, values=True, group='StageName', rankdir="RL", descend=False, text_width=25,
                 autoscale=None,
                 focus=True,
                 **kwargs):
        """

        :param model: a spanner (reference fragment)
        :param group: what to use for grouping. This can either be an attribute, or a callable function.
        """
        self._model = model.top()
        self._scenario = scenario
        self.values = values
        self.autoscale = autoscale

        self._text_width = int(text_width)
        self._group = group
        self._descend = descend

        self._graph = pydot.Dot(model.name, graph_type='digraph', strict=True, rankdir=rankdir, **kwargs)
        self._refflow = pydot.Node("ref flow", shape='none', height=0.25, label=model.flow['name'])
        self._graph.add_node(self._refflow)
        self.anchor(self._model, self._refflow, first=True)

    @property
    def graph(self):
        return self._graph

    @staticmethod
    def display():
        """
        To display a model graph from an ipython shell, run the following:

        from IPython.display import Image, display

        %matplotlib inline  # add if using jupyter

        def view_pydot(pdot):
            plt = Image(pdot.graph.create_png())
            display(plt)

        view_pydot(mg.graph)

        where mg is a ModelGraph instance.
        :return:
        """
        msg = '''\
# To display a model graph in a jupyter notebook environment, run the following
from IPython.display import Image, display

%matplotlib inline  # add if using jupyter

def view_pydot(pdot):
    plt = Image(pdot.graph.create_png())
    display(plt)
            
view_pydot(mg.graph)  # mg is a ModelGraph instance
'''
        print(msg)

    def get_group(self, frag):
        if self._group is None:
            return frag.name
        elif isinstance(self._group, str):
            return frag.get(self._group, '')
        elif callable(self._group):
            return self._group(frag)
        return frag.name

    def anchor(self, frag, node, first=False):
        _d = self._add_dir(frag, node)
        lbl = frag.uuid[:10] + '_anch'
        term = frag.termination(self._scenario)
        if term.is_null:
            cutoff = pydot.Node(lbl, shape="none", label=self._flowtext(frag.flow.name))
            self._graph.add_node(cutoff)
            self._graph.add_edge(pydot.Edge(cutoff, _d, arrowsize=0, minlen=1.3))
            return cutoff
        elif term.is_frag:
            if term.is_fg:
                if first:
                    anchor = pydot.Node(lbl, shape="rectangle", style="rounded", label=frag.get('Name', frag.name))
                else:
                    anchor = pydot.Node(lbl, shape="circle", label='')
                self._graph.add_node(anchor)
                self._graph.add_edge(pydot.Edge(anchor, _d, arrowsize=0))
                self.child_flows(frag, anchor)
            elif self._descend and frag.is_background:
                anchor = pydot.Node(lbl, shape="box", label=self._flowtext(frag.flow.name), height=0.25)
                self._graph.add_node(anchor)
                self._graph.add_edge(pydot.Edge(anchor, _d, arrowsize=0))
                self.child_flows(term.term_node, anchor)

            else:
                subfrag = pydot.Node(lbl, shape="box", height=0.25, label=self._flowtext(term.term_flow.name))
                self._graph.add_node(subfrag)
                self._graph.add_edge(pydot.Edge(subfrag, _d, arrowsize=0))
                self.child_flows(frag, subfrag)
        elif term.is_context:
            cx = pydot.Node(lbl, shape="box", style="dotted", height=0.25, label=self._flowtext(term.term_node.name))
            self._graph.add_node(cx)
            self._graph.add_edge(pydot.Edge(cx, _d, arrowsize=0, label=term.term_flow['name']))
            return cx

        elif term.is_process:
            anchor = pydot.Node(lbl, style="rounded", shape="rect", height=0.35,
                                label=line_break(term.term_node.name, start=self._text_width))
            self._graph.add_node(anchor)
            self._graph.add_edge(pydot.Edge(anchor, _d, arrow=False))
            self.child_flows(frag, anchor)

    def _add_dir(self, cf, node):
        _d = self._dir(cf)
        self._graph.add_edge(pydot.Edge(_d, node, arrowsize=0))
        return _d

    '''
    def _add_child(self, cf, node):
        cf_f = pydot.Node(cf.uuid[:10] + '_d', shape="box", height=0.25, label=cf.flow.name)
        self._graph.add_node(cf_f)
        self._graph.add_edge(pydot.Edge(cf_f, cf_d, arrowsize=False))
        return cf_f
    '''

    def child_flows(self, frag, node):
        for cf in sorted(frag.child_flows, key=lambda x: (x.direction, self.get_group(x), x.flow.unit, -x.observed_ev),
                         reverse=True):
            self.anchor(cf, node)

    def write_png(self, filename):
        self._graph.write_png(filename)

    def write_svg(self, filename):
        self._graph.write_svg(filename)



"""
lcia_eval.py

Generate information about consistency and completeness of LCI and LCIA methods.
"""
from antelope import ExchangeRef


class LciaEval(object):
    """
    The way this works is: we create two lists. One, of processes; the other, of LCIA methods.
    When we add a process to the list, we create a *unit mask* which has the same exchanges as the process's
    LCI, but all with unit magnitudes.  Then we create a "negative sort" which is initialized with all the contents of
    the unit mask.  Every time we add an LCIA method, we intersect the negative sort with the zeros of the resulting
    computation.  After all the lcia methods are added, what remains in the negative sort is the un-matched exchanges.
    Then we can look at those lists and decide what to do.
    """

    def __init__(self, unit_mask=False):
        self._rxs = []
        self._methods = []
        self._factors = dict()

        self._unit_mask = bool(unit_mask)
        self._lci = dict()  # maps process link to list of exchanges
        self._negative_sort = dict()  # maps process link to set of zero exchanges
        self._run = dict()

    def _run_lcia(self, rx, refresh=None):
        if len(self._methods):
            print('Running LCIA for %s' % rx.process.link)
        for q in self._methods:
            if (rx, q.link) in self._run:
                continue
            print(' -- LCIA %s: ' % q.link, end='')
            if rx.entity_type == 'fragment':
                res = rx.fragment_lcia(q, refresh=refresh).flatten()
            else:
                res = q.do_lcia(self._lci[rx], refresh=refresh)

            deets = list(res.details())
            print('%d hits, %d zeros; ' % (len(deets), len(list(res.zeros()))), end='')
            self._negative_sort[rx] &= set(res.zeros())
            print('%d remain' % len(self._negative_sort[rx]))
            self._run[rx, q.link] = res

    def _show_line(self, method, num=False):
        print('%30.30s ' % method.name, end='')
        for rx in self._rxs:
            if num:
                print(' %8.3g  ' % self._run[rx, method.link].total(), end='')
            else:
                print(' %8d  ' % len(list(self._run[rx, method.link].details())), end='')
        # I hate myself for taking the time to do this right now
        d = '  (%d)        ' % len(self._factors[method.link])
        print('%10.10s %s' % (d, method['Indicator'], ))

    def show(self, data=True):
        """
        damn he's really doing it
        :param data: True: report LCIA scores. False: report counts
        :return:
        """
        print('%30.30s ' % '', end='')
        for i in range(len(self._rxs)):
            print('   [%02d]     ' % i, end='')
        print()
        print('%30.30s ' % '', end='')
        for p in self._rxs:
            print('%-10.10s ' % p.process.origin, end='')
        print()
        print('%30.30s ' % '', end='')
        for p in self._rxs:
            print('%10.10s ' % p.process.name, end='')
        print()
        print('%30.30s ' % '', end='')
        for p in self._rxs:
            print('%10.10s ' % p.flow.name, end='')
        print()
        print('%30.30s ' % '', end='')
        for p in self._rxs:
            print('[%8.8s] ' % p.flow.unit, end='')
        print()

        print('%30.30s ' % '', end='')
        for p in self._rxs:
            print('   %5d   ' % len(self._lci[p]), end='')
        print()

        for k in self._methods:
            self._show_line(k, num=data)

        print('%30.30s ' % '', end='')
        for p in self._rxs:
            print('   %5d   ' % len(self._negative_sort[p]), end='')
        print()

    def recompute(self):
        self._run = dict()
        self._run_lcia_for_all_processes(refresh=True)

    def _run_lcia_for_all_processes(self, refresh=None):
        for p in self._rxs:
            self._run_lcia(p, refresh=refresh)

    def add_process(self, *rx_or_processes):
        for rx_or_p in rx_or_processes:
            if rx_or_p.entity_type == 'fragment':
                process = rx_or_p
                ios, internal = process.unit_flows()
                io = ios[0]
                rx = ExchangeRef(process, io.fragment.flow, io.fragment.direction, value=io.magnitude, is_reference=True)
                lci = [ExchangeRef(process, k.fragment.flow, k.fragment.direction, value=k.magnitude,
                                   termination=k.term.term_node) for k in internal if k.term.is_context]
            else:
                if rx_or_p.entity_type == 'process':
                    process = rx_or_p
                    rx = rx_or_p.reference()
                else:
                    rx = rx_or_p
                    process = rx.process

                if rx in self._rxs:
                    raise ValueError('Process already added')

                lci = list(process.lci(ref_flow=rx.flow))

            if self._unit_mask:
                # create a false inventory with 1 for everything
                lci = [ExchangeRef(x.process, x.flow, x.direction, value=1.0, termination=x.termination) for x in lci]
            self._lci[rx] = lci
            self._negative_sort[rx] = set(lci)
            self._rxs.append(rx)
            print('Adding Process %s: %d exchanges' % (process.link, len(set(lci))))
            self._run_lcia(rx)

    def add_lcia(self, *lcias):
        for lcia in lcias:
            if lcia in self._methods:
                raise ValueError('LCIA method already added')

            self._methods.append(lcia)
            self._factors[lcia.link] = list(lcia.factors())

            self._run_lcia_for_all_processes()

    '''
    Now, what outputs do we want?
    
    Flowables- we want to see all the flowables present, all the ones hit, all the ones missed
    make a table? lcia method by row, process by column- count hits, misses
    ultimately, though, we need to be curating flowables.   
    
    {we need an online qdb, is what we need}
    
    so this should be a tool for generating inputs to a curator- even if it is interactive- because how long would 
    it really take to clear the non-matched of 5100 flows? at 1 minute apiece that's like 2weeks of work.
    
    that's not what we would do though- we would load EXISTING lists of known syllables and then just mop up 
    the entries with only k terms, where k = 1, 2, ...
    we have all those lists from Rebe
    
    
    '''




from .randomizer import UniformRandomizer, LognormalRandomizer, DirichletRandomizer


class SimulationParameter:
    frag = None
    rand = None
    mean = 1.0

    def observe(self, n):
        self.frag.observe(self.rand.value(self.mean), scenario=n)

    def match(self, frag):
        return frag == self.frag


class UniformParameter(SimulationParameter):
    def __init__(self, fragment, low, high, scenario=None):
        """

        :param fragment: The fragment to be observed
        :param low: multiplicative lower bound for the uniform distribution
        :param high: multiplicative upper bound for the uniform distribution
        :param scenario: baseline scenario for observation
        """
        self.frag = fragment
        if scenario is None:
            self.mean = fragment.observed_ev
        else:
            self.mean = fragment.exchange_value(scenario)
        self.rand = UniformRandomizer(low, high)


class LognormalParameter(SimulationParameter):
    def __init__(self, fragment, shape, scenario=None):
        """

        :param fragment: The fragment to be observed
        :param shape: dimensionless shape parameter
        :param scenario: baseline scenario for observation
        """
        self.frag = fragment
        if scenario is None:
            self.mean = fragment.observed_ev
        else:
            self.mean = fragment.exchange_value(scenario)
        self.rand = LognormalRandomizer(shape)


class DirichletParameters:
    def __init__(self, *fragments, scale=1.0, alpha=None, scenario=None):
        """
        Implement a dirichlet distribution across the specified fragments
        :param fragments:
        :param scale: [1.0] Parameters will sum to this value
        :param alpha: [None] default: use uniform distribution parameters (e.g. [1, 1, 1]).
          Specify a single float to construct a distribution based on the fragments' exchange values, so that the
          parameters sum to alpha (e.g. 3)
          Specify a list or tuple of length equal to # of fragments to specify alpha explicitly.
        :param scenario: background scenario, used only when dynamically calculating alpha parameters
        """
        self.fragments = list(fragments)
        self.scale = scale
        if alpha is None:
            self.params = tuple([1] * len(self.fragments))
        else:
            if isinstance(alpha, list) or isinstance(alpha, tuple):
                if len(alpha) == len(self.fragments):
                    self.params = tuple(alpha)
                else:
                    raise ValueError('Bad alpha specification %s' % alpha)
            else:
                try:
                    a = float(alpha)
                    if scenario is None:
                        evs = [f.observed_ev for f in self.fragments]
                    else:
                        evs = [f.exchange_value(scenario) for f in self.fragments]
                    mul = a / sum(evs)
                    self.params = tuple(mul * ev for ev in evs)
                except TypeError:
                    raise ValueError('Bad alpha specification %s' % alpha)
        self.rand = DirichletRandomizer(*self.params)

    def observe(self, n):
        ds = self.rand.value(self.scale)
        for i, f in enumerate(self.fragments):
            f.observe(ds[i], scenario=n)

    def match(self, frag):
        return any(f == frag for f in self.fragments)



"""
Load flowable synonyms from an ecoinvent archive
"""
from synonym_dict import TermExists, MergeError


def load_ecoinvent_synonyms(cat, ecoinvent_origin, interface='exchange', merge_strategy=None):
    """
    this will not work when ecoinvent is accessed remotely-- entities_by_type would only get loaded flows...
    getting every flow will require at least one API call (to synonyms) for every flow already gotten...
    Better to get synonym info from a Qdb.

    :param cat:
    :param ecoinvent_origin: origin + interface that accesses a fully-complemented EcospoldV2Archive (with MasterData)
    :param interface: default 'exchange'
    :param merge_strategy: valid values: 'graft', 'prune', 'distinct', 'merge'.  None = use term manager default.
    :return: a list of flows that were not able to be loaded
    """
    broken = []
    count = 0
    ar = cat.get_archive(ecoinvent_origin, interface)
    if ar.__class__.__name__ == 'EcospoldV2Archive':
        # raise TypeError(ar, 'Wrong archive type')
        ar.load_flows()  # this loads synonyms
    else:
        print('Warning: unsupported archive type')
    for f in ar.entities_by_type('flow'):
        try:
            cat.lcia_engine.add_flow_terms(f, merge_strategy=merge_strategy)
            count += 1
        except (TermExists, MergeError) as e:
            print('broken (%s): %s' % (e.__class__.__name__, f.link))
            broken.append(f)
    print('loaded %d flows' % count)
    return broken




from .base import BaseTableOutput
from collections import defaultdict


class FlowablesGrid(BaseTableOutput):
    """
    This table can be produced dynamically

    """
    _near_headings = 'Flowable', 'Compartment'  # should be overridden
    _far_headings = 'Set',
    _returns_sets = True

    def _pull_row_from_item(self, item):
        """
        Items are either exchanges, factors, or flattened LciaResult components.
        The first has a 'flow' entity by attribute
        the middle has a flowable and also a context
        the last has an 'entity' which should be a flow.
        :param item:
        :return: 2-tuple: flowable name (string), compartment (Compartment)
        """
        if hasattr(item, 'entity_type'):
            if item.entity_type == 'exchange':
                return item.flow.name, item.termination
            elif item.entity_type == 'characterization':
                return item.flowable, item.context
        elif hasattr(item, 'entity'):
            flow = item.entity
            return flow.name, flow.context
        raise TypeError('Cannot find a flow in item of type %s' % type(item))

    def _generate_items(self, flow_collection):
        """
        Columns are: process ref, qty (lcia method) ref, lciaresult, or explicit iterator that provides exchanges,
        factors, or lciaresult components.

        :param flow_collection:
        :return:
        """
        if hasattr(flow_collection, 'flatten'):
            lcia = flow_collection.flatten()
            for component in lcia.components():
                if self._criterion(component.entity):
                    yield component
        elif hasattr(flow_collection, 'entity_type'):
            if flow_collection.entity_type == 'process':
                for exchange in flow_collection.inventory():
                    if self._criterion(exchange):
                        yield exchange
            elif flow_collection.entity_type == 'quantity':
                if flow_collection.is_lcia_method:
                    for cf in flow_collection.factors():
                        if self._criterion(cf):
                            yield cf
        else:
            # generic iterable-- assume exchange for now
            for x in flow_collection:
                if self._criterion(x):
                    yield x

    def _canonical(self, flow, unit=None):
        """

        use name: this is not quite right because parse_flow is too 'fuzzy'--
        add units to distinguish flows that get bypassed in convert()
        flowables because of CAS number; matches lots of different things that are all mapped to the same flowable (e.g.
        difference between fossil and biogenic CO2 is lost).  This is no longer WRONG because it adds instead of
        replacing.

        maybe that is the desirable course- but then I lose the ability to distinguish CO2 from peat, CO2 from gas, and
        co2 from land use change.  But maybe that is part of the meaning of a "flowables grid".

        still thinking.
        :param flow:
        :return:
        """
        name = self._qdb.get_flowable(flow).name
        unit = unit or flow.unit
        return '[%s] %s' % (unit, name)

    def _extract_data_from_item(self, objects):
        """
        We need to return dicts because we have _returns_sets = True.
        Items are either exchanges, factors, or flattened LciaResult components.
        First one is _value_dict (no existing public method to access)
        Second one is _locations (also no existing public method to access)
        Third one has no dict, it's just cumulative_result

        :param objects: dict item. list of objects
        :return:
        """
        d = defaultdict(float)
        if objects is None:
            return d
        for item in objects:
            if hasattr(item, 'entity_type'):
                if item.entity_type == 'exchange':
                    if hasattr(item, '_value_dict'):
                        for rx, v in item._value_dict.items():
                            key = '%s: %s' % (rx.direction, rx.flow['Name'])
                            d[key] += v
                    if item.value is not None:
                        d[self._canonical(item.flow)] += item.value
                elif item.entity_type == 'characterization':
                    # WHAT is going on here?
                    for loc, v in item._locations.items():
                        if loc == 'GLO':
                            key = self._canonical(item.flowable, unit=item.ref_quantity.unit)
                        else:
                            # there 'should not be' different cf values for the same flowable- but we know it happens
                            key = loc
                        if d[key] == 0:
                            d[key] = v
                        elif d[key] != v:
                            # right. very clever - log locales in canonical names
                            new_key = '[%s] %s' % (loc, item.flowable)
                            while d[new_key] != 0:
                                new_key += '.'
                            d[new_key] = v
                else:
                    raise TypeError('Unsure what to do with item %s' % item)
            elif hasattr(item, 'cumulative_result'):
                # print('%10.3g %s' % (item.cumulative_result, self._canonical(item.entity)))
                d[self._canonical(item.entity)] += item.cumulative_result
            else:
                raise TypeError('Unsure what to do with item %s' % item)
        return d

    def __init__(self, qdb, *args, include_flows=None, quell_locations=False):
        """

        :param qdb: Mandatory Qdb for resolving flowables and compartments
        :param args: supply process ref (exchanges), quantity ref (cfs), or lcia result (flows)
        :param include_flows: either None (include all flows) or an iterable of flowables to include
        :param quell_locations:  for characterizations, suppress reporting of locations other than GLO. [not yet impl.]
        """
        print('!!NOTE: _far_headings must be set')
        self._qdb = qdb
        if include_flows is not None:
            include_fbs = set([qdb.f_index(qdb.parse_flow(x)[0]) for x in include_flows])

            def criterion(x):
                return qdb.f_index(qdb.parse_flow(x)[0]) in include_fbs
        else:
            criterion = None

        super(FlowablesGrid, self).__init__(*args, criterion=criterion)



"""
Comparators

These tools enable quick comparison of sets of exchanges to see if they match
"""
from math import isclose


def _rx_key(rx):
    return rx.process.external_ref, rx.flow.external_ref, rx.direction


def _exch_key(exch):
    if exch.type == 'context':
        return exch.flow.external_ref, exch.direction, tuple(exch.termination)
    else:
        return exch.flow.external_ref, exch.direction, exch.termination


def compare_ref_exchanges(s1, s2):
    """
    Tests qualitative equivalence of two sets of *reference* exchanges (process, flow, direction).
    :param s1:
    :param s2:
    :return:
    """
    set1 = {_rx_key(s) for s in s1}
    set2 = {_rx_key(s) for s in s2}
    d1 = set1.difference(set2)
    d2 = set2.difference(set1)
    if len(d1) + len(d2) == 0:
        print('PASS (%d)' % len(set1))
    else:
        print('DIFFERENT (%d,%d / %d)' % (len(d1), len(d2), len(set1)))
    return d1, d2


def compare_exchanges(s1, s2):
    """
    Tests qualitative equivalence of two sets of *depenedent* exchanges (flow, direction, termination).
    Returns two sets: those present only in s1, those present only in s2
    :param s1:
    :param s2:
    :return: distinct exchanges in s1, distinct exchanges in s2. sum of lengths is 0 indicates success
    """
    set1 = {_exch_key(s) for s in s1}
    set2 = {_exch_key(s) for s in s2}
    d1 = set1.difference(set2)
    d2 = set2.difference(set1)
    if len(d1) + len(d2) == 0:
        print('PASS (%d)' % len(set1))
    else:
        print('DIFFERENT (%d,%d / %d)' % (len(d1), len(d2), len(set1)))
    return d1, d2


def compare_exchange_values(s1, s2, rel_tol=1e-6):
    """
    The exchanges are unordered, so it's nontrivial to test two sets of results against one another.
    We use the approach of mapping (flow.external_ref, direction, termination) to exchange value.
    This is a summary test that compares the two sets and returns the entries in s2 that do NOT match
    the entry in s1 (key error qualifies as nonmatch)
    Thus a return value of length 0 indicates a passed test.
    :param s1:
    :param s2:
    :param rel_tol: default 1e-6
    :return: a list of failed exchanges, expressed as 2-tuples: (bad_exch, good_val) or (bad_exch, None). 0-length
     return set indicates success
    """
    c = {_exch_key(x): x.value for x in s1}
    fail = []
    for s in s2:
        try:
            v = c[_exch_key(s)]
            if not isclose(v, s.value, rel_tol=rel_tol):
                fail.append((s, v))
        except KeyError:
            fail.append((s, None))
    if len(fail) == 0:
        print('PASS (%d)' % len(c))
    else:
        print('DIFFERENT (%d / %d)' % (len(fail), len(c)))
    return fail



import json
import os
import re
from antelope import UnknownOrigin


SYSTEM_MODELS = ('cutoff', 'apos', 'conseq', 'undefined')
LEVELS = ('low', 'medium', 'high')


class DataPersistenceError(Exception):
    pass


class _WrongNumberOfHits(Exception):
    pass


class EcoinventGrids(object):
    """
    Creates a special-purpose JSON file that maps grid geography, system model, and voltage level to external_ref
    for a given Ecoinvent version.
    """
    _version = None
    _file = None
    _j = None

    def _print(self, *args):
        print('%s: ' % self.__class__.__name__, end='')
        print(*args)

    def __init__(self, cat, version, path=None, levels=None, reset=False, local=None):
        """
        
        :param cat: Catalog that knows about ecoinvent 
        :param version: Ecoinvent semantic version available to the catalog (e.g. '3.8') 
        :param path: Directory to save + load JSON cache (default is the working directory)
        :param levels: which grid levels to use (default is all, i.e. ('low', 'medium', 'high'))
        :param reset: 
        :param local: [None] If present, prefix to pre-pend to the origin query (e.g. 'local.ecoinvent...')
        """
        self._cat = cat
        self._version = version
        self._levels = levels or LEVELS
        self._local = local
        filename = 'ecoinvent_grids_%s.json' % version
        if path is None:
            self._file = filename
        else:
            self._file = os.path.join(path, filename)
        if reset or not os.path.exists(self._file):
            self._j = {'data': 'EcoinventGrids',
                       'version': self._version}
            self.save()
        else:
            self.load()

    @property
    def version(self):
        return str(self._version)

    @property
    def models(self):
        for m in SYSTEM_MODELS:
            try:
                self._cat.query(self._org(m))
                yield m
            except UnknownOrigin:
                pass

    def _org(self, model):
        org = 'ecoinvent.%s.%s' % (self.version, model)
        if self._local:
            return '.'.join([str(self._local), org])
        else:
            return org

    @property
    def levels(self):
        for l in self._levels:
            yield l

    def load(self):
        with open(self._file, 'r') as fp:
            self._j = json.load(fp)
            try:
                assert self._j['data'] == 'EcoinventGrids', 'File content error %s' % self._j['data']
                assert self._j['version'] == self._version, 'File version error %s' % self._version
            except KeyError:
                raise DataPersistenceError('Keys missing: %s' % self._file)

    def save(self):
        with open(self._file, 'w') as fp:
            json.dump(self._j, fp, indent=2)

    def locations(self, loc, model=None):
        if model is None:
            model = next(self.models)

        return list(set(k['SpatialScope'] for k in self._cat.query(self._org(model)).processes(
            Name='^market( group)? for electricity',
            SpatialScope=loc)))  # should be an exact match

    def _get_em(self, _g, _m, _l):
        return list(_h.external_ref for _h in self._cat.query(self._org(_m)).processes(
            Name=('^market( group)? for electricity.*%s voltage' % _l),
            SpatialScope='^%s$' % _g))  # should be an exact match

    def _lookup_location(self, g):
        curr_g = self._j.get(g, None)
        new_g = {m: {l: self._get_em(g, m, l) for l in self.levels} for m in self.models}
        if len(new_g) == 0:
            raise ValueError('No valid models found')
        if curr_g == new_g:
            self._print('Location %s: no change' % g)
            return  # nothing to do
        else:
            if curr_g is None:
                self._print('adding location %s' % g)
            else:
                self._print('updating location %s' % g)
            self._j[g] = new_g
            self.save()

    def grids_by_location(self, loc, reset=False, show=False):
        """
        given a spatial scope, returns a dict of {level: {model: [grids]}} where level in self.levels and model in
        self.models
        :param loc:
        :param reset: [False]
        :param show: [False]
        :return:
        """
        if reset or (loc not in self._j):
            self._lookup_location(loc)
        if show:
            self.show_grids(loc)
        return self._j[loc]

    def show_grids(self, loc, model=None, level=None):
        if loc in self._j:
            for mdl, lvls in self._j[loc].items():
                if model:
                    if mdl != model:
                        continue
                for lvl, grids in lvls.items():
                    if level:
                        if lvl != level:
                            continue
                    for grid in grids:
                        t = self._cat.query(self._org(mdl)).get(grid)
                        print('%s: %s, %s: %s %s' % (loc, mdl, lvl, grid, t.name))

    def _create_grid_fragment(self, fg, name, model, grids, name_regex):
        origin = self._org(model)
        terms = [fg.cascade(origin).get(g) for g in grids]
        if name_regex is not None:
            terms = list(t for t in terms if bool(re.search(name_regex, t['name'], flags=re.IGNORECASE)))

        if len(terms) != 1:
            for t in terms:
                print(t)
            raise _WrongNumberOfHits(len(terms))
        term = terms[0]
        # keep the term's native reference flow for auto-consumption purposes, in case the fragment is expanded
        return fg.new_fragment(term.reference().flow, 'Output', external_ref=name, background=True, termination=term)

    @staticmethod
    def _add_to_chooser(fg, grid, level, scenario):
        # add to grid chooser
        cname = 'grid-chooser-%s' % level
        if fg[cname] is None:
            c = fg.new_fragment(fg['elec'], 'Output')
            fg.observe(c, name=cname)
        else:
            c = fg[cname]
        if scenario not in c.scenarios():
            c.terminate(grid, scenario=scenario)
            if c.term.is_null:  # set default to first encountered
                c.terminate(grid)

    def create_loc_grids(self, fg, loc, rshort=None, chooser=True, name_regex=None, suffix=None, show=False):
        """
        Builds grid models in the supplied foreground for the specified location, named 'grid-[region]-[model]-[level]',
        where 'region' is the first word of loc.

        If chooser is True, adds the grid as a termination to the 'grid-chooser-[level]' model under the scenario
        name '[region]-grid-[model]'. The grid chooser is created if it does not exist.

        :param fg: A foreground to store the grid models
        :param loc: Create grids for this locale. should be an exact match, e.g. ^(loc)$
        :param rshort: [None] short name to use for the grid in the scenario chooser (default is the first 'word' in
          the loc)
          (scenario will be `rshort`-grid-`model`; fragment will be grid-`rshort`-`model`-`level`)
        :param chooser: Whether to add the grids to a chooser fragment
        :param name_regex: Optional regex to filter from available grid process names [must hit exactly one]
        :param suffix: add dash-delimited suffix to fragment name, and scenario name if chooser is True
        :return:
        """
        fg.add_or_retrieve('elec', 'kWh', 'Electricity, at user')  # we need different levels, I suppose....
        md = self.grids_by_location(loc, show=show)
        if rshort is None:
            rshort = re.match('^\S+', loc).group(0)
        for model, ld in md.items():
            for level, grids in ld.items():
                if level not in self._levels:
                    continue
                name = 'grid-%s-%s-%s' % (rshort, model, level)
                if suffix:
                    name = '%s-%s' % (name, suffix)
                if fg[name] is None:
                    try:
                        grid = self._create_grid_fragment(fg, name, model, grids, name_regex)
                    except _WrongNumberOfHits as e:
                        if e.args[0] > 1:
                            print('Warning (skipped): Multiple hits [%d] for %s, %s, %s' % (e.args[0], loc, model, level))
                        else:
                            print('no target found for %s, %s, %s' % (loc, model, level))
                        continue

                else:
                    grid = fg[name]

                if chooser:
                    scenario = '%s-grid-%s' % (rshort, model)
                    if suffix:
                        scenario = '%s-%s' % (scenario, suffix)
                    self._add_to_chooser(fg, grid, level, scenario)



class ComponentsMixin(object):
    """
    A method to add the 'components' option to LcaModelRunner to write unit scores and node weights to a components
    file.,
    """
    components_headings = ['scenario', 'stage', 'node', 'node_weight', 'ref_units', 'method', 'category', 'indicator', 'units', 'unit_score']

    def _csv_format_components(self):
        return self.components_headings, self._gen_component_entries

    def _gen_component_entries(self, scenario, q, _rec=None, include_total=False, expand=False):
        if _rec is None:
            _rec = self._results[scenario, q]
        if include_total:
            yield self._gen_row(q, {
                'scenario': scenario,
                'node': 'Net Total',
                'node_weight': self._format(_rec.scale),
                'ref_units': 'scale',
                'unit_score': self._format(_rec.total() / _rec.scale)
            })
        for c in sorted(_rec.components(), key=lambda x: self._agg(x.entity)):
            if expand and not c.static:
                # recurse
                for y in self._gen_component_entries(q, scenario, _rec=c, expand=expand,
                                                     include_total=False): # don't print subtotals in recursion
                    yield y
            else:
                yield self._gen_row(q, {
                    'scenario': scenario,
                    'node': c.entity.name,
                    'stage': self._agg(c.entity),
                    'node_weight': self._format(c.node_weight),  # every component MUST be a summary because agg scores don't show up in traversals
                    'ref_units': c.entity.ref_unit,
                    'unit_score': self._format(c.unit_score)  #
                })





from pydantic import BaseModel


class LciaDefinition(BaseModel):
    origin: str
    external_ref: str
    method: str = 'LCIA Method'
    category: str = 'LCIA Category'
    name: str  # category
    unit: str
    indicator: str



import json

from antelope_reports.model_runner import LcaModelRunner


class QuickModelRunner(LcaModelRunner):

    @classmethod
    def from_csv_refs(cls, cat, fg_name, refs_file):
        """
        CSV file should
        :param cat:
        :param fg_name:
        :param refs_file:
        :return:
        """
        fg = cat.foreground(fg_name)

        with open(refs_file) as fp:
            refs = json.load(fp)

        refs = [ref.split('/', maxsplit=1) for ref in refs]
        procs = [cat.query(o).get(k) for o, k in refs]

        qm = cls(fg, procs)
        return qm

    def __init__(self, fg, refs=None, **kwargs):
        """

        :param fg: A foreground to store the fragment models
        :param refs: an iterable of process references to run.  If the process is multi-output, supply a 2-tuple of
        process_ref, ref_flow
        :param kwargs:
        """
        super(QuickModelRunner, self).__init__(**kwargs)

        self._fg = fg
        self._models = dict()

        if refs:
            for ref in refs:
                self.add_ref(ref)

    def add_ref(self, ref):
        if isinstance(ref, tuple):
            px = ref[0]
            rx = px.reference(ref[1])
        else:
            px = ref
            rx = px.reference()
        name = '%s (%s)' % (px.name, rx.flow.name)
        if name in self._models:
            print('Name %s is already registered' % name)
            return

        if px.entity_type == 'process':
            model = self._fg.create_process_model(px, ref_flow=rx.flow)
        else:
            model = px
        self._models[name] = model
        self.add_scenario(name)

    def _scenario_index(self, scenario):
        model = self._models[scenario]
        return model.name, model.flow.name, model.observed_ev, model.flow.unit

    def _run_scenario_lcia(self, scenario, lcia, **kwargs):
        model = self._models[scenario]
        return model.fragment_lcia(lcia, **kwargs)

    def compute_results(self, lcia):
        for l in lcia:
            self.run_lcia(l)

    def save_details(self, pg, prefix, lcia):
        node = self._models[pg].term.term_node.external_ref
        res = self.result(pg, lcia)
        fname = 'details-%s-%s-%s.json' % (prefix, node, lcia.external_ref)
        with open(fname, 'w') as fp:
            json.dump(res.serialize_components(detailed=True), fp, indent=2)


def generate_results(cat, lcia, fg_name, refs_file, out_file):
    qm = QuickModelRunner.from_csv_refs(cat, fg_name, refs_file)
    qm.compute_results(lcia)
    qm.scenario_summary_tbl(out_file)



"""
A common LCA use case: compare two different LCIA scores by stage

The interactive style:
 - identify the models
 - curate stages / groupings
 - build a model runner
 - spool out the results into a dataframe
 - send it here

"""

import pandas as pd
import numpy as np
from matplotlib.cm import get_cmap
from matplotlib import pyplot as plt
from math import ceil


def make_stack_plot_from_df(df, cases, f_u='result', filename=None, _qs=None, stage_colors=None, colormap='viridis',
                            cmap_factor=1.0, wspace=0.3, nrows=1):
    # Pivot data to calculate stacked bar components
    # we want to preserve the order in the dataframe
    if _qs is None:
        u_q = set(df['Quantity'])
        quantities = []
        units = []
        for i, k in enumerate(df['Quantity']):
            if k in u_q:
                quantities.append(k)
                u_q.remove(k)
                units.append(df['Unit'].loc[i])

            if len(u_q) == 0:
                break
    else:
        quantities = [q['ShortName'] for q in _qs]
        units = [q.unit for q in _qs]

    pivot_df = df.pivot_table(index=['Quantity', 'Case'], columns='Stage', values='Result', aggfunc='sum').fillna(
        0).reindex(
        pd.MultiIndex.from_product([quantities, cases], names=['Quantity', 'Case'])
    )

    # Unique indicators and stages
    stages = pivot_df.columns
    if stage_colors is None:
        _cm = get_cmap(colormap)
        stage_colors = {stage: _cm(i / len(stages) / cmap_factor) for i, stage in enumerate(stages)}

    # Create the figure with multiple axes
    # first, handle the subplot geometry
    n = min([nrows, len(quantities)])
    c = ceil(len(quantities) / n)

    fig, axes = plt.subplots(n, c, figsize=(5 * c, 6 * n), sharey=False, sharex=True)
    ax_legend = axes[0][-1]
    plt.subplots_adjust(wspace=wspace)

    if len(quantities) == 1:
        axes = [[axes]]  # Ensure axes is 2d array if only one plot
    else:
        axes = axes.reshape(1, n*c)  # this always returns a 2d array

    for ax in axes[0]:
        ax.set_axis_off()  # turn them all off

    # Loop through indicators to create subplots
    for ax, quantity, unit in zip(axes[0], quantities, units):
        # Filter data for this indicator
        indicator_data = pivot_df.loc[quantity]
        ax.set_axis_on()  # turn on when it's being used

        # Compute cumulative sums for stacking
        cumulative_sums = indicator_data.cumsum(axis=1).fillna(0)
        bar_positions = np.arange(len(cases))  # X positions for bars

        # Plot stacked bars
        for i, stage in enumerate(stages[::-1]):  # reverse direction for legend order
            ix = len(stages) - i - 2  # -1 for zero-indexing; -1 for cumsum offset
            bottom = cumulative_sums.iloc[:, ix] if ix >= 0 else None
            ax.bar(bar_positions, indicator_data[stage], label=stage, bottom=bottom, width=0.8,
                   color=stage_colors[stage])

        # Customize the axes
        ax.set_title(f"{f_u}: {quantity}")
        ax.set_xlabel(None)
        ax.set_ylabel(unit)
        ax.set_xlim([bar_positions[0] - 0.5, bar_positions[-1] + 0.5])
        ax.set_xticks(bar_positions)
        ax.set_xticklabels(cases, rotation=45)
        if ax is ax_legend:
            ax.legend(title="Stage", loc='upper left', bbox_to_anchor=(1, 1))

    # Set common y-label
    # fig.supylabel("Result")

    # Adjust layout
    # plt.tight_layout()
    if filename:
        plt.savefig(filename, bbox_inches='tight', dpi=300)
    return fig



from textwrap import wrap

import operator

from matplotlib import pyplot as plt
from matplotlib.container import BarContainer
from matplotlib.patches import Rectangle

USE_TEX = False

if USE_TEX:
    """
    matplotlib documentation lists valid TeX fonts
    https://matplotlib.org/stable/tutorials/text/usetex.html
    serif (\rmfamily):
        Computer Modern Roman, 
        Palatino (mathpazo), 
        Times (mathptmx), 
        Bookman (bookman), 
        New Century Schoolbook (newcent), 
        Charter (charter)

    sans-serif (\sffamily)"
        Computer Modern Serif, 
        Helvetica (helvet), 
        Avant Garde (avant)

    cursive (\rmfamily):
        Zapf Chancery (chancery)

    monospace (\ttfamily):
        Computer Modern Typewriter, 
        Courier (courier)
    """
    plt.rcParams.update({
        "text.usetex": True,
        "font.family": "serif",
        "font.sans-serif": "Helvetica",
        "font.serif": "Palatino"
    })
    plt.rc('text.latex', preamble=r'\usepackage[cm]{sfmath}')
else:
    """
    Get a list of valid font families like this:
    >>> from matplotlib.font_manager import findSystemFonts
    >>> flist = findSystemFonts()
    >>> for fname in flist:
            try:
                fm = matplotlib.font_manager.FontProperties(fname=fname)
                print(fm.get_name())
            except RuntimeError:
                print('## %s' % fname)
    >>>
    """
    plt.rcParams.update({"text.usetex": False,
                         "font.family": "P052",
                         })


class KPIBase(object):
    """
    Generates and saves (and/or shows?) an infographic chart with a graph panel and annotations.

    For now, we want the graphic to be something like the following:

    +--------------------------------------------------------------------+
    |INDICATOR                          |                                |
    | units                             |                                |
    |-----------------------------------|              chart             |
    |  YEAR                NUMBER  unit |                                |
    |                                   |                                |
    |___________________________________|                                |
    |comment                            |                                |
    +--------------------------------------------------------------------+


    """

    def _write_metric(self, record, amount, unit=None, fmt='%.1f', fontsize=None,
                      draw_symbol=None,
                      record_args=None,
                      amount_args=None):
        """
        Key operation-- pre-increments the row cursor and prints a statistical data point divided into
        record, amount, unit.
        :param record:
        :param amount:
        :param unit:
        :param fmt:
        :param fontsize:
        :param draw_symbol: [None] do nothing
          True: draw a legend for the current plot to the left of the record label
          'value': draw a legend for the current plot to the left of the amount
        :param record_args: dict of formatting kwargs to pass to record
        :param amount_args: dict of formatting kwargs to pass to amount+unit
        :return:
        """
        if record_args is None:
            record_args = dict()

        if amount_args is None:
            amount_args = dict()

        if fontsize is None:
            fontsize = self.fontsize

        self.content_row -= (fontsize * 1.15 / self.font_hgt)  # normalized units

        text_at = 0.015

        if draw_symbol:
            if isinstance(self.the_plot, BarContainer):
                patch = self.the_plot.patches[0]
                pps = {'linewidth': patch.get_linewidth(),
                       'edgecolor': patch.get_edgecolor(),
                       'facecolor': patch.get_facecolor()}
                text_at += 0.04
                mid = self.content_row + (fontsize * 0.1 / self.font_hgt)
                h = (fontsize * 0.7 / self.font_hgt)
                p = Rectangle((0.015, mid-0.02), 0.035, h, **pps)
                self.frame.add_patch(p)
            else:
                pps = {'linewidth': self.the_plot.get_linewidth(),
                       'linestyle': self.the_plot.get_linestyle(),
                       'color': self.the_plot.get_color(),
                       'marker': self.the_plot.get_marker(),
                       'markersize': self.the_plot.get_markersize(),
                       'markeredgecolor': self.the_plot.get_markeredgecolor(),
                       'markerfacecolor': self.the_plot.get_markerfacecolor()
                       }
                mid = self.content_row + (fontsize * 0.4 / self.font_hgt)
                ppy = (mid, mid)
                if draw_symbol == 'value':
                    amt_s = fmt % amount
                    amt_wid = len(amt_s) / 72  # hardcodes 16 chars per inch x 6 inch width bc I'm dumb/lazy
                    ppx = (self.content_f[0] - amt_wid - 0.06, self.content_f[0] - amt_wid - 0.015)
                    self.frame.plot(ppx, ppy, **pps)
                else:
                    text_at += 0.06
                    self.frame.plot((0.015, 0.06), ppy, **pps)

        t0 = self.frame.text(text_at, self.content_row, '%s' % record, fontsize=fontsize,
                             ha='left', va='baseline', **record_args)
        t1 = self.frame.text(self.content_f[0], self.content_row, fmt % amount,
                             ha='right', va='baseline', **amount_args)
        if not unit:
            unit = '%s' % self.unit
        if unit == '%':
            unit = self.pct()

        if unit == 'none':
            t2 = None  # suppress unit
        else:
            t2 = self.frame.text(self.content_f[1], self.content_row, unit,
                                 ha='left', va='baseline', **amount_args)
        self.metrics.append((t0, t1, t2))

    def write_text(self, text, fontsize=None, **kwargs):
        if fontsize is None:
            fontsize = self.fontsize

        self.content_row -= (fontsize * 1.15 / self.font_hgt)  # normalized units

        t0 = self.frame.text(0.015, self.content_row, text, fontsize=fontsize,
                             ha='left', va='baseline', **kwargs)

        self.metrics.append((t0, None, None))

    def __init__(self, title, subtitle, unit, comment=None, figsize=(6.5, 1.8), text_f=0.55, frame_lw=1.2, long=False,
                 graph_l_m_in=0.39, graph_b_m_in=0.39,
                 metrics=(0.85, 0.88),
                 title_in=0.5, comment_in=0.55, fontsize=12):

        # data elements
        self._xs = []
        self._ys = []
        self._the_plots = []

        # display elements
        self._series = -1
        self._this = -1
        self.unit = unit  # data unit

        self.fontsize = fontsize

        wid, hgt = figsize
        self.font_hgt = 72 * hgt  # figure height in font points

        self.text_f = text_f
        self.title_f = 1.0 - title_in / hgt
        self.frame_lw = frame_lw

        comment_f = comment_in / hgt

        self.title_pin = 1.0 - 0.33 / hgt  # vertical pin for title/subtitle break - fixed
        if long:
            self.numchars = int(16 * wid)  # approx width of content in number of characters (at 10 pt)
        else:
            self.numchars = int(16 * wid * text_f)  # approx width of content in number of characters (at 10 pt)

        self.content_row = self.title_f - 0.035
        self.content_f = [text_f * k for k in metrics]

        self.metrics = []

        self.f = f = plt.figure(figsize=figsize)

        # content axes
        self.frame = f.add_axes((0, 0, 1, 1))
        plt.axis('tight')

        # chart axes
        graph_l_m = graph_l_m_in / wid  # left hand margin for y labels (in)
        graph_b_m = graph_b_m_in / hgt  # lower margin for x labels (in)

        if long:
            graph_b = comment_f + graph_b_m
        else:
            graph_b = graph_b_m
        self.chart = f.add_axes((text_f + graph_l_m, graph_b,
                                 0.99 - text_f - graph_l_m, 0.92 - graph_b))

        # title
        self.title = self.frame.text(0.015, self.title_pin + 0.015, title, fontsize=self.fontsize * 1.6,
                                     ha='left', va='bottom')
        self.subtitle = self.frame.text(0.015, self.title_pin, '%s (%s)' % (subtitle, self.unit), fontsize=self.fontsize,
                                        ha='left', va='top')

        self.comment = self.frame.text(0.015, comment_f - 0.035, 'comment', ha='left', va='top',
                                       fontsize=self.fontsize * 0.83,
                                       visible=False)
        if long:
            self._comment_frame = self.frame.plot((0, 1), (comment_f, comment_f), color='k',
                                                  linewidth=self.frame_lw, visible=True)[0]
        else:
            self._comment_frame = self.frame.plot((0, self.text_f), (comment_f, comment_f), color='k',
                                                  linewidth=self.frame_lw * 0.7, visible=False)[0]

        self._format_frame(long, comment_f)
        self._format_chart()
        self.set_comment(comment)  # None will pass

    @property
    def the_plot(self):
        return self._the_plots[self._series]

    @property
    def N(self):
        return len(self._xs)

    @property
    def n(self):
        return len(self.ys)

    @property
    def xs(self):
        return self._xs[self._series]

    @property
    def ys(self):
        return self._ys[self._series]

    @property
    def this_x(self):
        return self.xs[self._this]

    @property
    def this_y(self):
        return self.ys[self._this]

    def _format_frame(self, long, comment_f):
        self.frame.set_xlim([0, 1])
        self.frame.set_ylim([0, 1])
        self.frame.set_xticks(())
        self.frame.set_yticks(())

        # draw frames
        self.frame.plot((0, self.text_f), (self.title_f, self.title_f), color='k', linewidth=self.frame_lw)
        if long:
            self.frame.plot((self.text_f, self.text_f), (comment_f, 1), color='k', linewidth=self.frame_lw)
        else:
            self.frame.plot((self.text_f, self.text_f), (0, 1), color='k', linewidth=self.frame_lw)

    def _format_chart(self):
        self.chart.spines['right'].set_visible(False)
        self.chart.spines['left'].set_visible(False)
        self.chart.spines['top'].set_visible(False)
        self.chart.spines['bottom'].set_visible(False)
        self.chart.set_clip_on(False)

        # what a terrible syntactic trick
        self.baseline, = self.chart.plot((0, 1), (0, 0), linestyle='-', color='k', linewidth=0.7, visible=False)

        # def _draw_frame(self):

    def rescale(self, g_gap):
        self.chart.autoscale()  # rescale the axes to extents

        xlim = self.chart.get_xlim()
        xlim = [xlim[0] - g_gap, xlim[1] + g_gap]
        self.chart.set_xlim(xlim)
        self.baseline.set_xdata(xlim)
        self.baseline.set_visible(True)
        self.baseline.set_clip_on(False)

        ylim = self.chart.get_ylim()
        if ylim[1] < 0:
            ylim = [ylim[0], 0]
        elif ylim[0] > 0:
            ylim = [0, ylim[1]]
        self.chart.set_ylim(ylim)
        # set ticklabel font sizes
        self.chart.tick_params(axis='x', labelsize=self.fontsize * 0.83)
        self.chart.tick_params(axis='y', labelsize=self.fontsize * 0.83)

    def highlight(self, this=-1, marker='o', hi_color=(0.1, 0.1, 0.5), hi_size=10, a_fmt='%.1f', jog=12):
        """

        :param this: [-1] index of plot point to highlight
        :param marker:
        :param hi_color: color of highlight marker
        :param hi_size: size of highlight marker
        :param a_fmt: ['%.1f'] printf format for highlight annotation
        :param jog: [12] vertical position control for annotation (pts)
        :return:
        """
        self._this = this
        if self._this is not None:
            # highlight 'this' value
            self.chart.plot(self.this_x, self.this_y, linestyle='none', marker=marker, markerfacecolor='none',
                            markeredgecolor=hi_color, markersize=hi_size, clip_on=False)

            ylim = self.chart.get_ylim()

            # annotate 'this' value
            # put below if value is too high
            if (ylim[1] - self.this_y) / (ylim[1] - ylim[0]) < 0.18:
                jog *= -1.3

            # omit unit, since it's printed in the frame
            self.chart.annotate(a_fmt % self.this_y, xy=(self.this_x, self.this_y),
                                xytext=(0, jog), textcoords='offset points',
                                ha='center')

    def draw_plot(self, xs, ys, linestyle='-', marker='o', color=(0.5, 0.1, 0.1), g_gap=0.38, fill=None,
                  **kwargs):
        """
        Add a dataset to the graphic as a line plot. xs and ys will be set as the current data.
        'this' indicates which value to hilight (default -1, set to None to disable hilight)
        :param xs: x coordinates
        :param ys: y coordinates
        :param linestyle: data linestyle
        :param marker: data marker
        :param color: data line color
        :param g_gap: [0.38] padding to add to the left and right of the x axis
        :param fill: if True, fill the space between the line and (x axis, or prior dataset). If int, fill the space
         between the line and the indicated dataset
        :param kwargs:
        :return:
        """
        #####
        self._xs.append(list(xs))
        self._ys.append(list(ys))
        # def draw_plot(self):

        if fill is not None:
            if fill is True:
                yb = 0
            else:
                try:
                    fill = int(fill)
                    yb = self._ys[fill]
                except (TypeError, IndexError):
                    yb = 0
            self.chart.fill_between(self.xs, self.ys, yb, color=color, alpha=0.3)

        # line plot with markers
        pl, = self.chart.plot(self.xs, self.ys, linestyle=linestyle, marker=marker, color=color, clip_on=False, **kwargs)
        self._the_plots.append(pl)

        self.rescale(g_gap)

    def draw_barplot(self, xs, ys, color=(0.121, 0.467, 0.706), g_gap=0.38, **kwargs):
        """

        :param xs:
        :param ys:
        :param color:
        :param g_gap:
        :param kwargs:
        :return:
        """
        self._xs.append(list(xs))
        self._ys.append(list(ys))

        bar = self.chart.bar(self.xs, self.ys, color=color, **kwargs)
        self._the_plots.append(bar)

        self.rescale(g_gap)

    def set_comment(self, comment):
        if comment:
            self._comment_frame.set_visible(True)
            cw = '\n'.join(wrap(comment, self.numchars))
            self.comment.set_visible(True)
            self.comment.set_text(cw)

    def add_to_comment(self, add):
        if add:
            t = self.comment.get_text()
            self.comment.set_text(t + '\n' + add)

    def unset_comment(self):
        self._comment_frame.set_visible(False)
        self.comment.set_visible(False)

    def set_title(self, title, units):
        if title:
            self.title.set_text(title)
        if units:
            self.subtitle.set_text('%s (%s)' % (units, self.unit))

    def save(self, filename):
        plt.figure(self.f)
        plt.savefig(filename)

    @staticmethod
    def pct():
        if plt.rcParams['text.usetex']:
            return '\%'
        else:
            return '%'

    def report_value(self, ix=None, draw_label=None, **kwargs):
        """

        :param ix:
        :param draw_label: prepend text: [True] plot label, if specified, ['str']: the string
        :param kwargs:
        :return:
        """
        if ix:
            x = self.xs[ix]
            y = self.ys[ix]
        else:
            x = self.this_x
            y = self.this_y
        if draw_label:
            if draw_label is True:
                lbl = self.the_plot.get_label()
            else:
                lbl = str(draw_label)
            metric = '%s, %s' % (lbl, x)
        else:
            metric = 'Value for %s' % x
        self._write_metric(metric, y, **kwargs)

    def report_change(self, ix, pct=True, **kwargs):
        prior_x = self.xs[ix]
        prior_y = self.ys[ix]
        metric = "Change since %s" % prior_x

        if pct:
            ch = (self.this_y - prior_y) * 100 / prior_y
            self._write_metric(metric, ch, '%', fmt='%+.1f', **kwargs)
        else:
            ch = (self.this_y - prior_y) / prior_y
            self._write_metric(metric, ch, '', fmt='%+.3f', **kwargs)

    def _label_metric(self, default):
        lbl = self.the_plot.get_label()
        if lbl.startswith('_'):  # not set by user
            return default
        return '%s, %s' % (default, lbl)

    def report_total(self, metric=None, weight=None, **kwargs):
        if weight:
            metric = self._label_metric(metric or 'Weighted Sum')
            # python-native inner product
            y = sum(map(operator.mul, self.ys, weight))
        else:
            metric = self._label_metric(metric or 'Total')
            y = sum(self.ys)
        self._write_metric(metric, y, **kwargs)

    def report_average(self, metric=None, weight=None, **kwargs):
        if weight:
            metric = self._label_metric(metric or 'Weighted Average')
            # python-native inner product
            y = sum(map(operator.mul, self.ys, weight)) / sum(weight)
        else:
            metric = self._label_metric(metric or 'Simple Average')
            y = sum(self.ys) / len(self.ys)
        self._write_metric(metric, y, **kwargs)

    def report_literal(self, metric, amount, unit, **kwargs):
        """
        Report value not derived from current data
        :param metric:
        :param amount:
        :param unit:
        :param kwargs:
        :return:
        """
        self._write_metric(metric, amount, unit=unit, **kwargs)

    def content(self, **kwargs):
        pass

    def run(self, title=None, subtitle=None, comment=None, filename=None, **kwargs):
        self.set_title(title, subtitle)
        self.set_comment(comment)
        self.content(**kwargs)
        if filename:
            self.save(filename)


class ExtensiveKpi(KPIBase):
    def content(self, weight=None, metric=None, **kwargs):
        self.report_value(**kwargs)
        for k in [-2, 0]:
            self.report_change(k)
        self.report_total(weight=weight, metric=metric, **kwargs)


class IntensiveKpi(KPIBase):
    def content(self, weight=None, metric=None, **kwargs):
        self.report_value(**kwargs)
        for k in [-2, 0]:
            self.report_change(k)
        self.report_average(weight=weight, metric=metric, **kwargs)



from antelope import EntityNotFound


def __make_link(fg, parent, models, child):
    if isinstance(child, tuple):
        f = []
        if isinstance(parent, list):
            for i, p in enumerate(parent):
                if p is None or child[i] is None:
                    f.append(None)
                else:
                    term = models.get(child[i])
                    f.append(fg.new_fragment(term.flow, 'Output', parent=p))
                    f[-1].terminate(term)
        else:
            for c in child:
                term = models.get(c)
                f.append(fg.new_fragment(term.flow, 'Output', parent=parent))
                f[-1].terminate(term)
    else:
        term = models.get(child)
        if parent is None:
            f = fg.new_fragment(term.flow, 'Input')
        else:
            f = fg.new_fragment(term.flow, 'Output', parent=parent)
        f.terminate(term)
    return f


def _make_route(fg, models, route):
    """
    Cycle through
    :param fg:
    :param route:
    :return:
    """
    parent = first = None
    for step in route:
        if first is None and isinstance(step, tuple):
            raise ValueError('Model spec cannot begin with a tuple! %s' % step)
        child = __make_link(fg, parent, models, step)
        if first is None:
            first = child
        parent = child
    return first


def make_routes(study, models, routes):
    """
    Disposition routes get built from scratch during LCA runtime in LCA study foreground.  Components are drawn from
    models foreground.

    This is a bit of a Kluge, but here is the idea:

    A ROUTE SPECIFICATION consists of a sequence of STAGES (as a tuple).  Each STAGE is EITHER:
     - a fragment present in the models foreground, OR
     - a BRANCH.  Any given route specification may contain AT MOST ONE BRANCH. The branch is implemented as a tuple
       of fragments.  Each subsequent stage must be a tuple with the same length.  The first stage may not be a branch.

    For each stage, the route builder will retrieve the target and create a fragment whose flow matches the target's.
    It will then terminate the fragment to the target.  That stage then becomes the parent of the subsequent stage.

    When a branch is encountered, it will create multiple child flows from the parent, one for each entry in the branch.
    EVERY SUBSEQUENT STAGE must have the same number of entries as the branch, and these entries are appended as
    child nodes to the branched parents.  Use None to end a branch (None must be supplied as a placeholder throughout
    the route).

    :param study: LcForeground to contain the dynamically constructed routes
    :param models: LcForeground that has the modeling content used in the routes.  Both foregrounds may be the same.
    :param routes: a dict of name : route specification, where the name is assigned as the external_ref
    :return: a list of names (external_refs) of the created fragments, should be identical to the keys of routes.
    """
    refs = []
    for k, v in routes.items():
        refs.append(k)
        if study[k] is None:
            route = _make_route(study, models, v)
            study.observe(route, name=k)
    return refs


def _check_p_map(fg, models, p_map):
    nc = 0
    for k, v in p_map.items():
        try:
            fg.get(k)
        except EntityNotFound:
            try:
                models.get(k)
            except EntityNotFound:
                print('Product map entry %s not found' % k)
                return False
        if v is None:
            nc += 1
        else:
            try:
                float(v)
            except (ValueError, AttributeError):
                print('Key %s has non-floatable value %s' % (k, v))
                return False
    if nc != 1:
        print('Wrong number of balance flows (%d)' % nc)
        return False
    return True


def build_market_mix(fg, parent, models, p_map, freight=False):
    if freight:
        sense = 'Input'
    else:
        sense = 'Output'
    if _check_p_map(fg, models, p_map):
        for k, v in p_map.items():
            try:
                term = fg.get(k)
            except EntityNotFound:
                term = models.get(k)
            if freight:
                flow = term
            else:
                flow = parent.flow
            if v is None:
                c = fg.new_fragment(flow, sense, parent=parent, balance=True)
            else:
                c = fg.new_fragment(flow, sense, parent=parent, value=v)
                mix_name = 'mix-%s-%s' % (parent.flow.external_ref, k)
                fg.observe(c, name=mix_name, exchange_value=v)
            if freight:
                # don't terminate freight flows
                continue
            c.terminate(term, term_flow = term.flow)
            if term.external_ref == 'frag-t_waste':
                c.term.descend = False
                c['StageName'] = 'Waste to Landfill'

        return True
    return False


def market_mixers(study):
    return {k.external_ref: k for k in study.fragments(show_all=True) if k.external_ref.startswith('mix')}



"""
One of the most common interpretive tools for LCA is to express the impacts of one activity in terms of
another. The Equivalency is a formalization of that.  The user supplies a reference exchange which is the comparison
activity, and a "measure" which is an optional dependent exchange for which the unit comparison is drawn (default:
just use the reference flow), and also documentary information (title and docstring).

The equivalency can then be queried by supplying an LCIA quantity, which returns the total score per unit of measure.
"""
from collections import namedtuple

EquivSpec = namedtuple('EquivSpec', ('title', 'origin', 'external_ref', 'measure', 'docstring'))  # measure = "how many (unit)"


class Equivalency:
    """
    The purpose of this is to create an easy way to compare two different activities in terms of their relative
    LCIA scores.
    """

    @classmethod
    def from_spec(cls, cat, e_spec: EquivSpec, ref_flow=None):
        """
        The EquivSpec provides a concise programmatic way of specifying equivalencies.
        :param cat:
        :param e_spec:
        :param ref_flow:
        :return:
        """
        rx = cat.query(e_spec.origin).get(e_spec.external_ref).reference(ref_flow)
        return cls(e_spec.title, rx, measure=e_spec.measure, docstring=e_spec.docstring)

    def __init__(self, title, rx, measure=None, docstring=None, **kwargs):
        """

        :param title: short title
        :param rx: Reference exchange
        :param measure: Dependent exchange for unit comparison [if None, uses the reference]
        :param docstring: annotation text
        :param kwargs: not currently used
        """
        self.title = title
        self._rx = rx
        self._meas = measure

        if measure is None:
            self._mx = self.rx
            self._measure = self._e.reference_value(self.ref_flow)
        else:
            self._mx = next(self._e.exchange_values(flow=measure))
            self._measure = self._e.exchange_relation(self.ref_flow, self._mx.flow, self._mx.direction)

        self.docstring = docstring or ''
        self.init_args = kwargs

    @property
    def _e(self):
        return self._rx.process

    @property
    def unit(self):
        return self._mx.flow.unit

    @property
    def measure(self):
        return self._measure

    @property
    def rx(self):
        return self._rx

    @property
    def ref_flow(self):
        return self.rx.flow

    @property
    def origin(self):
        return self._e.origin

    @property
    def external_ref(self):
        return self._e.external_ref

    def query_lcia(self, quantity):
        res = quantity.do_lcia(self._e.lci(ref_flow=self.rx))
        return res.total() / self.measure


class EquivalencyGenerator:
    """
    A handy tool for a limited set of circumstances.  Allows a user to retain a collection of LCIA quantities
    mapped to specific equivalencies.
    """
    def __init__(self, cat):
        self.cat = cat
        self.qs = []
        self._entries = dict()

    def add_equiv(self, q, equiv_spec):
        equiv = Equivalency.from_spec(self.cat, equiv_spec)
        if not hasattr(q, 'entity_type'):
            q = self.cat.get_canonical(q)
        if q not in self.qs:
            self.qs.append(q)
        self._entries[q.external_ref] = equiv

    def check_equiv(self, q):
        if q in self.qs:
            return self._entries[q.external_ref]

    @property
    def entries(self):
        for k in sorted(self._entries.keys()):
            yield k



"""
Let's try this again.
"""
from antelope import EntityNotFound, ForegroundInterface
from ..fg_builder.lca_study import LcaStudy


class NestedLcaStudy(LcaStudy):
    """
    An improved formalization of the nested LCA study model first devised for CATRA Phase 1.  The basic idea is that
    the core of the model is provided by an MFA fragment whose inventory produces inputs and outputs.  The study
    fragment (which is EPHEMERAL) contains 3 levels-- to each level are added child flows that handle inputs and
    outputs of the MFA fragment, according to a mapping that can be pre-specified or ad hoc.

    This class has three main components:
     - Initialization - creates the "superstructure" of reference_flow <- study <- logistics <- activity <- #DATA#
       The #DATA# fragment is the user input.  When traversed, it will produce an inventory of cut-off flows that get
       automatically linked to child flows in the superstructure.  These child flows are grouped by their reference
       quantity:

         ref qty  |  superstructure node
         ---------+---------------------
         mass     |  study_container
         freight  |  logistics_container
         activity |  activity_container

       Flows with other reference quantities can also be terminated, most likely in the study_container.

     - Route Construction methods - used to build the components to which superstructure child flows are linked.  These
       fragments are all dynamically created (no statefulness to worry about) and stored in the study foreground.

     - Model populating methods - user-facing methods to attach data fragments to the study.

    For each tier
    """
    @classmethod
    def create_from_names(cls, cat, fg_name, models_name, data_name=None, **kwargs):
        """
        In the event that the models + data foregrounds already exist, this is a convenience function to
        encapsulate the foreground query calls.
        :param cat:
        :param fg_name:
        :param models_name:
        :param data_name:
        :param kwargs:
        :return:
        """
        fg = cat.foreground(fg_name, create=True, reset=True)
        models = cat.foreground(models_name)
        if data_name:
            data = cat.foreground(data_name)
        else:
            data = None
        return cls(fg, models=models, data=data, **kwargs)

    def __init__(self, foreground, models=None, data=None, reference_flow='MFA-LCA Study',
                 study_container='study_container',
                 logistics_container='logistics_container',
                 activity_container='activity_container'):
        """

        :param foreground: foreground query - to contain the study model - ephemeral
        :param models: [optional] foreground query to contain upstream models
        :param data: foreground query for MFA data
        :param reference_flow: flow name for the study's reference flow
        :param study_container: [eponymous] external_ref
        :param logistics_container: [eponymous] external_ref
        :param activity_container: [eponymous] external_ref
        """
        super(NestedLcaStudy, self).__init__(foreground, models, reference_flow=reference_flow)

        self._data = data

        self._study = study_container
        self._logis = logistics_container
        self._activ = activity_container

    @property
    def data(self):
        return self._data

    def set_data_fg(self, data_fg):
        if not isinstance(data_fg, ForegroundInterface):
            raise TypeError(data_fg)
        self._data = data_fg

    def _resolve_term(self, term):
        """
        Grab an entry from the foreground[s] associated with the project: first _fg, then _models, then _data if
        it exists. Store successful lookups in a "phone book" to cut down on queries doomed to 404
        :param term:
        :return:
        """
        try:
            return super(NestedLcaStudy, self)._resolve_term(term)
        except EntityNotFound:
            if self.data is not None:
                t = self._data.get(term)
                self._phonebook[term] = t
                return t
            raise

    def _matching_foreground(self, origin):
        try:
            return super(NestedLcaStudy, self)._matching_foreground(origin)
        except KeyError:
            if self.data is not None:
                if origin == self._data.origin:
                    return self._data
            raise

    @property
    def study_container(self):
        try:
            return self._fg.get(self._study)
        except EntityNotFound:
            study = self._fg.new_fragment(self.reference_flow, 'Output', name=self._ref)
            self._fg.observe(study, name=self._study, termination=self.logistics_container)
        return self._fg.get(self._study)

    @property
    def logistics_container(self):
        try:
            return self._fg.get(self._logis)
        except EntityNotFound:
            study = self._fg.new_fragment(self.reference_flow, 'Output', name='%s_logistics' % self._ref)
            self._fg.observe(study, name=self._logis, termination=self.activity_container)
        return self._fg.get(self._logis)

    @property
    def activity_container(self):
        try:
            return self._fg.get(self._activ)
        except EntityNotFound:
            study = self._fg.new_fragment(self.reference_flow, 'Output', name='%s_nodes' % self._ref)
            self._fg.observe(study, name=self._activ)
        return self._fg.get(self._activ)

    def install_observation_model(self, prov_frag, scope=None):
        if scope is None:
            scope = prov_frag.get('Scope')
            if scope is None:
                raise ValueError('Must provide scope!')
        self.activity_container.clear_termination(scope)
        self.activity_container.terminate(prov_frag, scope)

    '''
    def add_logistics_route(self, flow, provider, descend=False, **kwargs):
        """

        :param flow:
        :param provider:
        :param descend: [False] whether to traverse [True] or aggregate [False] subfragments
        :param kwargs: StageName='Logistics', other args to fragment
        :return:
        """
        try:
            return next(self.logistics_container.children_with_flow(flow))
        except StopIteration:
            c = self.fg.new_fragment(flow, 'Input', parent=self.logistics_container, **kwargs)
            c.terminate(self._resolve_term(provider), descend=descend)
            return c
    '''
